# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT
import numpy as np
import os

from .counterfactual_group import CounterfactualGroup
from ..dataset_configs.config_map import scperturb_datasets, scperturb_gene_datasets


class GenePerturbGroup(CounterfactualGroup):
    """Create GenePerturbGroup Group Class object. This is for single-cell gene perturbation prediction tasks benchmark.

    Args:
            path (str, optional): the path to store/retrieve the GenePerturb group datasets.
    """

    _GENE_COLS = [
        'UMI_count', 'cancer', 'cell_line', 'disease', 'guide_id', 'ncounts',
        'ngenes', 'nperts', 'organism', 'percent_mito', 'percent_ribo',
        'perturbation', 'perturbation_type', 'tissue_type'
    ]

    def __init__(self,
                 dataset="scperturb_gene_ReplogleWeissman2022_k562_essential",
                 path="./data",
                 file_format="csv"):
        """Create a GenePerturbGroup prediction benchmark group class."""
        self.name = "GenePerturbGroup"
        self.path = os.path.join(path, self.name)
        self.dataset_names = [
            "scperturb_gene_NormanWeissman2019",
            "scperturb_gene_ReplogleWeissman2022_rpe1",
            "scperturb_gene_ReplogleWeissman2022_k562_essential"
        ]
        self.file_format = file_format

        from ..multi_pred.perturboutcome import PerturbOutcome
        dataset = dataset
        assert dataset in self.dataset_names, "{} dataset not in {}".format(
            dataset, self.dataset_names)
        data = PerturbOutcome(dataset)
        self.split = data.get_split()
        self.test = self.get_test()

    def get_train_valid_split(self):
        """parameters included for compatibility. this benchmark has a fixed train/test split."""
        train = self.split[0]["train"]
        val = self.split[0]["dev"]
        self.ctrl = np.array(
            np.mean(train[train.obs.condition == 'ctrl'].X,
                    axis=0)).reshape(-1,)
        return train, val

    def get_test(self):
        return self.split[0]["test"]

    def evaluate(self, pred):
        # pred is a dictionary with perturbation name in the format of X+X, where single gene is X+ctrl
        from scipy.stats import pearsonr
        test = self.test
        ctrl = self.ctrl
        mean_truth = {}
        for condition in test.obs.condition.unique():
            out = test[test.obs.condition == condition].X
            mean_truth[condition] = np.array(np.mean(out, axis=0)).reshape(-1,)

        pearson_delta = []
        for i, j in mean_truth.items():
            pearson_delta.append(pearsonr(pred[i] - ctrl, j - ctrl)[0])

        pearson_delta = np.mean(pearson_delta)

        A = test.var.gene_name.values
        de_20 = {
            i.split('_')[1]:
                np.array(
                    [np.where(A == b)[0][0] if b in A else -1 for b in j[:20]])
            for i, j in test.uns['rank_genes_groups_cov_all'].items()
        }
        mse_20_de = []
        for i, j in mean_truth.items():
            mse_20_de.append(np.mean((pred[i][de_20[i]] - j[de_20[i]])**2))
        mse_20_de = np.mean(mse_20_de)

        direction = []
        for i, j in mean_truth.items():
            direction.append(
                sum(
                    np.sign(pred[i][de_20[i]] -
                            ctrl[de_20[i]]) == np.sign(j[de_20[i]] -
                                                       ctrl[de_20[i]])) / 20)
        direction = np.mean(direction)

        return {
            'pearson_delta_all_genes': pearson_delta,
            'mse_20_de_genes': mse_20_de,
            'ratio_direction_correct_20_de_genes': direction
        }



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import numpy as np
from typing import List

try:
    from rdkit import Chem, DataStructs
    from rdkit.Chem import AllChem
    from rdkit import rdBase

    rdBase.DisableLog("rdApp.error")
    from rdkit.Chem.Fingerprints import FingerprintMols
    from rdkit.Chem import MACCSkeys
except:
    raise ImportError(
        "Please install rdkit by 'conda install -c conda-forge rdkit'! ")

from ...utils import print_sys
from ..oracle.oracle import (
    smiles_to_rdkit_mol,
    smiles_2_fingerprint_ECFP4,
    smiles_2_fingerprint_FCFP4,
    smiles_2_fingerprint_AP,
    smiles_2_fingerprint_ECFP6,
)
from ._smiles2pubchem import smiles2pubchem


def canonicalize(smiles):
    mol = Chem.MolFromSmiles(smiles)
    if mol is not None:
        return Chem.MolToSmiles(mol, isomericSmiles=True)
    else:
        return None


def smiles2morgan(s, radius=2, nBits=1024):
    """Convert smiles into Morgan Fingerprint.

    Args:
      smiles: str
      radius: int (default: 2)
      nBits: int (default: 1024)

    Returns:
      fp: numpy.array

    """
    try:
        s = canonicalize(s)
        mol = Chem.MolFromSmiles(s)
        features_vec = AllChem.GetMorganFingerprintAsBitVect(mol,
                                                             radius,
                                                             nBits=nBits)
        features = np.zeros((1,))
        DataStructs.ConvertToNumpyArray(features_vec, features)
    except:
        print_sys("rdkit not found this smiles for morgan: " + s +
                  " convert to all 0 features")
        features = np.zeros((nBits,))
    return features


def smiles2rdkit2d(s):
    """Convert smiles into 200-dim Normalized RDKit 2D vector.

    Args:
      smiles: str

    Returns:
      fp: numpy.array

    """

    s = canonicalize(s)
    try:
        from descriptastorus.descriptors import rdDescriptors, rdNormalizedDescriptors
    except:
        raise ImportError(
            "Please install pip install git+https://github.com/bp-kelley/descriptastorus and pip install pandas-flavor"
        )
    try:
        generator = rdNormalizedDescriptors.RDKit2DNormalized()
        features = np.array(generator.process(s)[1:])
        NaNs = np.isnan(features)
        features[NaNs] = 0
    except:
        print_sys("descriptastorus not found this smiles: " + s +
                  " convert to all 0 features")
        features = np.zeros((200,))
    return np.array(features)


def smiles2daylight(s):
    """Convert smiles into 2048-dim Daylight feature.

    Args:
      smiles: str

    Returns:
      fp: numpy.array

    """
    try:
        s = canonicalize(s)
        NumFinger = 2048
        mol = Chem.MolFromSmiles(s)
        bv = FingerprintMols.FingerprintMol(mol)
        temp = tuple(bv.GetOnBits())
        features = np.zeros((NumFinger,))
        features[np.array(temp)] = 1
    except:
        print_sys("rdkit not found this smiles: " + s +
                  " convert to all 0 features")
        features = np.zeros((2048,))
    return np.array(features)


def smiles2maccs(s):
    """Convert smiles into maccs feature.

    Args:
      smiles: str

    Returns:
      fp: numpy.array

    """
    s = canonicalize(s)
    mol = Chem.MolFromSmiles(s)
    fp = MACCSkeys.GenMACCSKeys(mol)
    arr = np.zeros((0,), dtype=np.float64)
    DataStructs.ConvertToNumpyArray(fp, arr)
    return arr


"""
  ECFP2 ---- 1
  ECFP4 ---- 2
  ECFP6 ---- 3
  xxxxxxxxx  ------  https://github.com/rdkit/benchmarking_platform/blob/master/scoring/fingerprint_lib.py 

"""


def smiles2ECFP2(smiles):
    """Convert smiles into ECFP2 Morgan Fingerprint.

    Args:
      smiles: str

    Returns:
      fp: rdkit.DataStructs.cDataStructs.UIntSparseIntVect

    """
    nbits = 2048
    smiles = canonicalize(smiles)
    molecule = smiles_to_rdkit_mol(smiles)
    fp = AllChem.GetMorganFingerprintAsBitVect(molecule, 1, nBits=nbits)
    arr = np.zeros((0,), dtype=np.float64)
    DataStructs.ConvertToNumpyArray(fp, arr)
    return arr


def smiles2ECFP4(smiles):
    """Convert smiles into ECFP4 Morgan Fingerprint.

    Args:
      smiles: str

    Returns:
      fp: rdkit.DataStructs.cDataStructs.UIntSparseIntVect

    """
    nbits = 2048
    smiles = canonicalize(smiles)
    molecule = smiles_to_rdkit_mol(smiles)
    fp = AllChem.GetMorganFingerprintAsBitVect(molecule, 2, nBits=nbits)
    arr = np.zeros((0,), dtype=np.float64)
    DataStructs.ConvertToNumpyArray(fp, arr)
    return arr


def smiles2ECFP6(smiles):
    """Convert smiles into ECFP6 Morgan Fingerprint.

    Args:
      smiles: str, a SMILES string

    Returns:
      fp: rdkit.DataStructs.cDataStructs.UIntSparseIntVect

    refer: https://github.com/rdkit/benchmarking_platform/blob/master/scoring/fingerprint_lib.py

    """
    nbits = 2048
    smiles = canonicalize(smiles)
    molecule = smiles_to_rdkit_mol(smiles)
    fp = AllChem.GetMorganFingerprintAsBitVect(molecule, 3, nBits=nbits)
    arr = np.zeros((0,), dtype=np.float64)
    DataStructs.ConvertToNumpyArray(fp, arr)
    return arr


# def smiles2smart(smiles):


class MoleculeFingerprint:
    """
    Example:
    MolFP = MoleculeFingerprint(fp = 'ECFP6')
    out = MolFp('Clc1ccccc1C2C(=C(/N/C(=C2/C(=O)OCC)COCCN)C)\C(=O)OC')
    # np.array([1, 0, 1, .....])
    out = MolFp(['Clc1ccccc1C2C(=C(/N/C(=C2/C(=O)OCC)COCCN)C)\C(=O)OC',
                'CCCOc1cc2ncnc(Nc3ccc4ncsc4c3)c2cc1S(=O)(=O)C(C)(C)C'])
    # np.array([[1, 0, 1, .....],
                [0, 0, 1, .....]])

    Supporting FPs:
    Basic_Descriptors(atoms, chirality, ....), ECFP2, ECFP4, ECFP6, MACCS, Daylight-type, RDKit2D, Morgan, PubChem
    """

    def __init__(self, fp="ECFP4"):
        fp2func = {
            "ECFP2": smiles2ECFP2,
            "ECFP4": smiles2ECFP4,
            "ECFP6": smiles2ECFP6,
            "MACCS": smiles2maccs,
            "Daylight": smiles2daylight,
            "RDKit2D": smiles2rdkit2d,
            "Morgan": smiles2morgan,
            "PubChem": smiles2pubchem,
        }
        try:
            assert fp in fp2func
        except:
            raise Exception("The fingerprint you specify are not supported. \
              It can only among 'ECFP2', 'ECFP4', 'ECFP6', 'MACCS', 'Daylight', 'RDKit2D', 'Morgan', 'PubChem'"
                           )

        self.fp = fp
        self.func = fp2func[fp]

    def __call__(self, x):
        if type(x) == str:
            return self.func(x)
        elif type(x) == list:
            lst = list(map(self.func, x))
            arr = np.vstack(lst)
            return arr


def smiles2selfies(smiles):
    """Convert smiles into selfies.

    Args:
      smiles: str, a SMILES string

    Returns:
      selfies: str, a SELFIES string.

    """
    smiles = canonicalize(smiles)
    return sf.encoder(smiles)


def selfies2smiles(selfies):
    """Convert selfies into smiles.

    Args:
      selfies: str, a SELFIES string.

    Returns:
      smiles: str, a SMILES string

    """
    return canonicalize(sf.decoder(selfies))


def smiles2mol(smiles):
    """Convert SMILES string into rdkit.Chem.rdchem.Mol.

    Args:
      smiles: str, a SMILES string.

    Returns:
      mol: rdkit.Chem.rdchem.Mol

    """
    smiles = canonicalize(smiles)
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return None
    Chem.Kekulize(mol)
    return mol


def bondtype2idx(bond_type):
    if bond_type == Chem.rdchem.BondType.SINGLE:
        return 1
    elif bond_type == Chem.rdchem.BondType.DOUBLE:
        return 2
    elif bond_type == Chem.rdchem.BondType.TRIPLE:
        return 3
    elif bond_type == Chem.rdchem.BondType.AROMATIC:
        return 4


def smiles2graph2D(smiles):
    """convert SMILES string into two-dimensional molecular graph feature

    Args:
      smiles, str, a SMILES string

    Returns:
      idx2atom: dict, map from index to atom's symbol, e.g., {0:'C', 1:'N', ...}
      adj_matrix: np.array

    """
    smiles = canonicalize(smiles)
    mol = smiles2mol(smiles)
    n_atoms = mol.GetNumAtoms()
    idx2atom = {atom.GetIdx(): atom.GetSymbol() for atom in mol.GetAtoms()}
    adj_matrix = np.zeros((n_atoms, n_atoms), dtype=int)
    for bond in mol.GetBonds():
        a1 = bond.GetBeginAtom()
        a2 = bond.GetEndAtom()
        idx1 = a1.GetIdx()
        idx2 = a2.GetIdx()
        bond_type = bond.GetBondType()
        bond_idx = bondtype2idx(bond_type)
        adj_matrix[idx1, idx2] = bond_idx
        adj_matrix[idx2, idx1] = bond_idx
    return idx2atom, adj_matrix


def get_mol(smiles):
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return None
    Chem.Kekulize(mol)
    return mol


############### PyG begin ###############
ELEM_LIST = [
    "C",
    "N",
    "O",
    "S",
    "F",
    "Si",
    "P",
    "Cl",
    "Br",
    "Mg",
    "Na",
    "Ca",
    "Fe",
    "Al",
    "I",
    "B",
    "K",
    "Se",
    "Zn",
    "H",
    "Cu",
    "Mn",
    "unknown",
]
ATOM_FDIM = len(ELEM_LIST) + 6 + 5 + 4 + 1
BOND_FDIM = 5 + 6
MAX_NB = 6
# https://github.com/kexinhuang12345/DeepPurpose/blob/master/DeepPurpose/chemutils.py


def onek_encoding_unk(x, allowable_set):
    if x not in allowable_set:
        x = allowable_set[-1]
    return list(map(lambda s: x == s, allowable_set))


def get_atom_features(atom):
    return torch.Tensor(
        onek_encoding_unk(atom.GetSymbol(), ELEM_LIST) +
        onek_encoding_unk(atom.GetDegree(), [0, 1, 2, 3, 4, 5]) +
        onek_encoding_unk(atom.GetFormalCharge(), [-1, -2, 1, 2, 0]) +
        onek_encoding_unk(int(atom.GetChiralTag()), [0, 1, 2, 3]) +
        [atom.GetIsAromatic()])


def smiles2PyG(smiles):
    """convert SMILES string into torch_geometric.data.Data

    Args:
      smiles, str, a SMILES string

    Returns:
      data, torch_geometric.data.Data

    """
    smiles = canonicalize(smiles)
    mol = Chem.MolFromSmiles(smiles)
    n_atoms = mol.GetNumAtoms()
    atom_features = [get_atom_features(atom) for atom in mol.GetAtoms()]
    atom_features = torch.stack(atom_features)
    y = [atom.GetSymbol() for atom in mol.GetAtoms()]
    y = list(
        map(
            lambda x: ELEM_LIST.index(x)
            if x in ELEM_LIST else len(ELEM_LIST) - 1, y))
    y = torch.LongTensor(y)
    bond_features = []
    for bond in mol.GetBonds():
        a1 = bond.GetBeginAtom()
        a2 = bond.GetEndAtom()
        idx1 = a1.GetIdx()
        idx2 = a2.GetIdx()
        bond_features.extend([[idx1, idx2], [idx2, idx1]])
    bond_features = torch.LongTensor(bond_features)

    data = Data(x=atom_features, edge_index=bond_features.T)

    return data


def molfile2PyG(molfile):
    smiles = molfile2smiles(molfile)
    smiles = canonicalize(smiles)
    return smiles2PyG(smiles)


############### PyG end ###############


############### DGL begin ###############
def smiles2DGL(smiles):
    """convert SMILES string into dgl.DGLGraph

    Args:
      smiles, str, a SMILES string

    Returns:
      g: dgl.DGLGraph()

    """
    smiles = canonicalize(smiles)
    mol = Chem.MolFromSmiles(smiles)
    n_atoms = mol.GetNumAtoms()
    bond_features = []
    for bond in mol.GetBonds():
        a1 = bond.GetBeginAtom()
        a2 = bond.GetEndAtom()
        idx1 = a1.GetIdx()
        idx2 = a2.GetIdx()
        bond_features.extend([[idx1, idx2], [idx2, idx1]])
    src, dst = tuple(zip(*bond_features))
    g = dgl.DGLGraph()
    g.add_nodes(n_atoms)
    g.add_edges(src, dst)
    return g


############### DGL end ###############

from ._xyz2mol import xyzfile2mol


def mol2smiles(mol):
    smiles = Chem.MolToSmiles(mol)
    smiles = canonicalize(smiles)
    return smiles


def xyzfile2smiles(xyzfile):
    """convert xyzfile into smiles string.

    Args:
      xyzfile: str, file

    Returns:
      smiles: str, a SMILES string

    """
    mol, _ = xyzfile2mol(xyzfile)
    smiles = mol2smiles(mol)
    smiles = canonicalize(smiles)
    return smiles


def xyzfile2selfies(xyzfile):
    """convert xyzfile into SELFIES string.

    Args:
      xyzfile: str, file

    Returns:
      selfies: str, a SELFIES string.

    """
    smiles = xyzfile2smiles(xyzfile)
    smiles = canonicalize(smiles)
    selfies = smiles2selfies(smiles)
    return selfies


def distance3d(coordinate_1, coordinate_2):
    return np.sqrt(
        sum([(c1 - c2)**2 for c1, c2 in zip(coordinate_1, coordinate_2)]))


def upper_atom(atomsymbol):
    return atomsymbol[0].upper() + atomsymbol[1:]


def xyzfile2graph3d(xyzfile):
    atoms, charge, xyz_coordinates = read_xyz_file(file)
    num_atoms = len(atoms)
    distance_adj_matrix = np.zeros((num_atoms, num_atoms))
    for i in range(num_atoms):
        for j in range(i + 1, num_atoms):
            distance = distance3d(xyz_coordinates[i], xyz_coordinates[j])
            distance_adj_matrix[i, j] = distance_adj_matrix[j, i] = distance
    idx2atom = {
        idx: upper_atom(str_atom(atom)) for idx, atom in enumerate(atoms)
    }
    mol, BO = xyzfile2mol(xyzfile)
    return idx2atom, distance_adj_matrix, BO


############## end xyz2mol ################


def sdffile2smiles_lst(sdffile):
    """convert SDF file into a list of SMILES string.

    Args:
      sdffile: str, file

    Returns:
      smiles_lst: a list of SMILES strings.

    """
    from rdkit.Chem.PandasTools import LoadSDF

    df = LoadSDF(sdffile, smilesName="SMILES")
    smiles_lst = df["SMILES"].to_list()
    return smiles_lst


def sdffile2mol_conformer(sdffile):
    """convert sdffile into a list of molecule conformers.

    Args:
      sdffile: str, file

    Returns:
      smiles_lst: a list of molecule conformers.

    """
    from rdkit.Chem.PandasTools import LoadSDF

    df = LoadSDF(sdffile, smilesName="SMILES")
    mol_lst = df["ROMol"].tolist()
    conformer_lst = []
    for mol in mol_lst:
        conformer = mol.GetConformer(id=0)
        conformer_lst.append(conformer)
    mol_conformer_lst = list(zip(mol_lst, conformer_lst))
    return mol_conformer_lst


def mol_conformer2graph3d(mol_conformer_lst):
    """convert list of (molecule, conformer) into a list of 3D graph.

    Args:
      mol_conformer_lst: list of tuple (molecule, conformer)

    Returns:
      graph3d_lst: a list of 3D graph.
            each graph has (i) idx2atom (dict); (ii) distance_adj_matrix (np.array); (iii) bondtype_adj_matrix (np.array)

    """
    graph3d_lst = []
    bond2num = {"SINGLE": 1, "DOUBLE": 2, "TRIPLE": 3, "AROMATIC": 4}
    for mol, conformer in mol_conformer_lst:
        atom_num = mol.GetNumAtoms()
        distance_adj_matrix = np.zeros((atom_num, atom_num))
        bondtype_adj_matrix = np.zeros((atom_num, atom_num), dtype=int)
        idx2atom = {i: v.GetSymbol() for i, v in enumerate(mol.GetAtoms())}
        positions = []
        for i in range(atom_num):
            pos = conformer.GetAtomPosition(i)
            coordinate = np.array([pos.x, pos.y, pos.z]).reshape(1, 3)
            positions.append(coordinate)
        positions = np.concatenate(positions, 0)
        for i in range(atom_num):
            for j in range(i + 1, atom_num):
                distance_adj_matrix[i,
                                    j] = distance_adj_matrix[j, i] = distance3d(
                                        positions[i], positions[j])
        for bond in mol.GetBonds():
            a1 = bond.GetBeginAtom().GetIdx()
            a2 = bond.GetEndAtom().GetIdx()
            bt = bond.GetBondType()
            bondtype_adj_matrix[a1, a2] = bond2num[str(bt)]
            bondtype_adj_matrix[a1, a2] = bond2num[str(bt)]
        graph3d_lst.append((idx2atom, distance_adj_matrix, bondtype_adj_matrix))
    return graph3d_lst


def sdffile2graph3d_lst(sdffile):
    """convert SDF file into a list of 3D graph.

    Args:
      sdffile: SDF file

    Returns:
      graph3d_lst: a list of 3D graph.
            each graph has (i) idx2atom (dict); (ii) distance_adj_matrix (np.array); (iii) bondtype_adj_matrix (np.array)

    """
    mol_conformer_lst = sdffile2mol_conformer(sdffile)
    graph3d_lst = mol_conformer2graph3d(mol_conformer_lst)
    return graph3d_lst


def sdffile2selfies_lst(sdf):
    """convert sdffile into a list of SELFIES strings.

    Args:
      sdffile: str, file

    Returns:
      selfies_lst: a list of SELFIES strings.

    """
    smiles_lst = sdffile2smiles_lst(sdf)
    selfies_lst = list(map(smiles2selfies, smiles_lst))
    return selfies_lst


def smiles_lst2coulomb(smiles_lst):
    """convert a list of SMILES strings into coulomb format.

    Args:
      smiles_lst: a list of SELFIES strings.

    Returns:
      features: np.array

    """
    molecules = [Molecule(smiles, "smiles") for smiles in smiles_lst]
    for mol in molecules:
        mol.to_xyz(optimizer="UFF")
    cm = CoulombMatrix(cm_type="UM", n_jobs=-1)
    features = cm.represent(molecules)
    features = features.to_numpy()
    return features
    ## (nmol, max_atom_n**2),
    ## where max_atom_n is maximal number of atom in the smiles_lst
    ## features[i].reshape(max_atom_n, max_atom_n)[:3,:3]  -> 3*3 Coulomb matrix


def sdffile2coulomb(sdf):
    """convert sdffile into a list of coulomb feature.

    Args:
      sdffile: str, file

    Returns:
      coulomb feature: np.array
    """
    smiles_lst = sdffile2smiles_lst(sdf)
    return smiles_lst2coulomb(smiles_lst)


def xyzfile2coulomb(xyzfile):
    smiles = xyzfile2smiles(xyzfile)
    smiles = canonicalize(smiles)
    return smiles_lst2coulomb([smiles])


# 2D_format = ['SMILES', 'SELFIES', 'Graph2D', 'PyG', 'DGL', 'ECFP2', 'ECFP4', 'ECFP6', 'MACCS', 'Daylight', 'RDKit2D', 'Morgan', 'PubChem']
# 3D_format = ['Graph3D', 'Coulumb']


## XXX2smiles
def molfile2smiles(molfile):
    """convert molfile into SMILES string

    Args:
      molfile: str, a file.

    Returns:
      smiles: str, SMILES strings

    """
    mol = Chem.MolFromMolFile(molfile)
    smiles = Chem.MolToSmiles(mol)
    smiles = canonicalize(smiles)
    return smiles


def mol2file2smiles(molfile):
    """convert mol2file into SMILES string

    Args:
      mol2file: str, a file.

    Returns:
      smiles: str, SMILES strings

    """
    mol = Chem.MolFromMol2File(molfile)
    smiles = Chem.MolToSmiles(mol)
    smiles = canonicalize(smiles)
    return smiles


## smiles2xxx

atom_types = ["C", "N", "O", "H", "F", "unknown"]  ### Cl, S?


def atom2onehot(atom):
    """convert atom to one-hot feature vector

    Args:
      'C'

    Returns:
      [1, 0, 0, 0, 0, ..]

    """
    onehot = np.zeros((1, len(atom_types)))
    idx = atom_types.index(atom)
    onehot[0, idx] = 1
    return onehot


def atomstring2atomfeature(atom_string_list):
    atom_features = [atom2onehot(atom) for atom in atom_string_list]
    atom_features = np.concatenate(atom_features, 0)
    return atom_features


def raw3D2pyg(raw3d_feature):
    """convert raw3d feature to pyg (torch-geometric) feature

    Args:
      raw3d_feature: (atom_string_list, positions, y)
        - atom_string_list: list, each element is an atom, length is N
        - positions: np.array, shape: (N,3)
        - y: float

    Returns:
      data = Data(x=x, pos=pos, y=y)

    """
    import torch
    from torch_geometric.data import Data  ### global

    # atom_string_list, positions, y = raw3d_feature
    atom_string_list, positions = raw3d_feature
    atom_features = atomstring2atomfeature(atom_string_list)
    atom_features = torch.from_numpy(atom_features)
    positions = torch.from_numpy(positions)
    # y = torch.FloatTensor(y)
    # data = Data(x = atom_features, pos = positions, y = y)
    data = Data(x=atom_features, pos=positions)
    return data


convert_dict = {
    "SMILES": [
        "SELFIES",
        "Graph2D",
        "PyG",
        "DGL",
        "ECFP2",
        "ECFP4",
        "ECFP6",
        "MACCS",
        "Daylight",
        "RDKit2D",
        "Morgan",
        "PubChem",
    ],
    "SELFIES": [
        "SMILES",
        "Graph2D",
        "PyG",
        "DGL",
        "ECFP2",
        "ECFP4",
        "ECFP6",
        "MACCS",
        "Daylight",
        "RDKit2D",
        "Morgan",
        "PubChem",
    ],
    "mol": [
        "SMILES",
        "SELFIES",
        "Graph2D",
        "PyG",
        "DGL",
        "ECFP2",
        "ECFP4",
        "ECFP6",
        "MACCS",
        "Daylight",
        "RDKit2D",
        "Morgan",
        "PubChem",
    ],
    "mol2": [
        "SMILES",
        "SELFIES",
        "Graph2D",
        "PyG",
        "DGL",
        "ECFP2",
        "ECFP4",
        "ECFP6",
        "MACCS",
        "Daylight",
        "RDKit2D",
        "Morgan",
        "PubChem",
    ],
    "SDF": ["SMILES", "SELFIES", "Graph3D", "Coulumb"],
    "XYZ": ["SMILES", "SELFIES", "Graph3D", "Coulumb"],
    "Raw3D": ["PyG3D"],
}

fingerprints_list = [
    "ECFP2",
    "ECFP4",
    "ECFP6",
    "MACCS",
    "Daylight",
    "RDKit2D",
    "Morgan",
    "PubChem",
]

twoD_format = [
    "SMILES",
    "SELFIES",
    "mol",
    "mol2",
]
threeD_format = [
    "SDF",
    "XYZ",
    "PyG3D",
    "Raw3D",
    "distance",
    "Coulumb",
    "shape",
]  ### shape:mesh


class MolConvert:
    """MolConvert: convert the molecule from src formet to dst format.




    Example:
        convert = MolConvert(src = ‘SMILES’, dst = ‘Graph2D’)
        g = convert(‘Clc1ccccc1C2C(=C(/N/C(=C2/C(=O)OCC)COCCN)C)\C(=O)OC’)
        # g: graph with edge, node features
        g = convert(['Clc1ccccc1C2C(=C(/N/C(=C2/C(=O)OCC)COCCN)C)\C(=O)OC',
                  'CCCOc1cc2ncnc(Nc3ccc4ncsc4c3)c2cc1S(=O)(=O)C(C)(C)C'])
        # g: a list of graphs with edge, node features
        if src is 2D, dst can be only 2D output
        if src is 3D, dst can be both 2D and 3D outputs
        src: 2D - [SMILES, SELFIES]
              3D - [SDF file, XYZ file]
        dst: 2D - [2D Graph (+ PyG, DGL format), Canonical SMILES, SELFIES, Fingerprints]
              3D - [3D graphs (adj matrix entry is (distance, bond type)), Coulumb Matrix]
    """

    def __init__(self, src="SMILES", dst="Graph2D", radius=2, nBits=1024):
        self._src = src
        self._dst = dst
        self._radius = radius
        self._nbits = nBits

        self.convert_dict = convert_dict
        if "SELFIES" == src or "SELFIES" == dst:
            try:
                import selfies as sf

                global sf
            except:
                raise Exception(
                    "Please install selfies via 'pip install selfies'")

        if "Coulumb" == dst:
            try:
                from chemml.chem import CoulombMatrix, Molecule

                global CoulombMatrix, Molecule
            except:
                raise Exception(
                    "Please install chemml via 'pip install pybel' and 'pip install chemml'. "
                )

        if "PyG" == dst:
            try:
                import torch
                from torch_geometric.data import Data

                global torch
                global Data
            except:
                raise Exception(
                    "Please install PyTorch Geometric via 'https://pytorch-geometric.readthedocs.io/en/latest/notes/installation.html'."
                )

        if "DGL" == dst:
            try:
                import dgl

                global dgl
            except:
                raise Exception("Please install DGL via 'pip install dgl'.")

        try:
            assert src in self.convert_dict
        except:
            raise Exception("src format is not supported")
        try:
            assert dst in self.convert_dict[src]
        except:
            raise Exception("It is not supported to convert src to dst.")

        if src in twoD_format:
            ### 1. src -> SMILES
            if src == "SMILES":
                f1 = canonicalize
            elif src == "SELFIES":
                f1 = selfies2smiles
            elif src == "mol":
                f1 = molfile2smiles
            elif src == "mol2":
                f1 = mol2file2smiles

            ### 2. SMILES -> all
            # 'SMILES', 'SELFIES', 'Graph2D', 'PyG', 'DGL', 'ECFP2', 'ECFP4', 'ECFP6', 'MACCS', 'Daylight', 'RDKit2D', 'Morgan', 'PubChem'
            if dst == "SMILES":
                f2 = canonicalize
            elif dst == "SELFIES":
                f2 = smiles2selfies
            elif dst == "Graph2D":
                f2 = smiles2graph2D
            elif dst == "PyG":
                f2 = smiles2PyG
            elif dst == "DGL":
                f2 = smiles2DGL
            elif dst == "ECFP2":
                f2 = smiles2ECFP2
            elif dst == "ECFP4":
                f2 = smiles2ECFP4
            elif dst == "ECFP6":
                f2 = smiles2ECFP6
            elif dst == "MACCS":
                f2 = smiles2maccs
            elif dst == "Daylight":
                f2 = smiles2daylight
            elif dst == "RDKit2D":
                f2 = smiles2rdkit2d
            elif dst == "Morgan":
                f2 = smiles2morgan
            elif dst == "PubChem":
                f2 = smiles2pubchem
            self.func = lambda x: f2(f1(x))
        elif src in threeD_format:
            pass

        ### load from xyz file, input is a filename (str), only contain one smiles
        if src == "XYZ" and dst == "SMILES":
            self.func = xyzfile2smiles
        elif src == "XYZ" and dst == "SELFIES":
            self.func = xyzfile2selfies
        elif src == "XYZ" and dst == "Graph3D":
            self.func = xyzfile2graph3d
        elif src == "XYZ" and dst == "Coulumb":
            self.func = xyzfile2coulomb

        ### SDF file
        elif src == "SDF" and dst == "Graph3D":
            self.func = sdffile2graph3d_lst
        elif src == "SDF" and dst == "SMILES":
            self.func = sdffile2smiles_lst
        elif src == "SDF" and dst == "SELFIES":
            self.func = sdffile2selfies_lst
        elif src == "SDF" and dst == "Coulumb":
            self.func = sdffile2coulomb
        elif src == "Raw3D" and dst == "PyG3D":
            self.func = raw3D2pyg

    def __call__(self, x):
        if type(x) == np.ndarray:
            x = x.tolist()

        if type(x) == str:
            if self.func != smiles2morgan:
                return self.func(x)
            else:
                return self.func(x, radius=self._radius, nBits=self._nbits)
        elif type(x) == list:
            if self.func != smiles2morgan:
                out = list(map(self.func, x))
            else:
                lst = []
                for x0 in x:
                    lst.append(
                        self.func(x0, radius=self._radius, nBits=self._nbits))
                out = lst
            if self._dst in fingerprints_list:
                out = np.array(out)
            return out

    @staticmethod
    def eligible_format(src=None):
        """
        given a src format, output all the available format of the src format
        Example
        MoleculeLink.eligible_format('SMILES')
        ## ['Graph', 'SMARTS', ...]
        """
        if src is not None:
            try:
                assert src in convert_dict
            except:
                raise Exception("src format is not supported")
            return convert_dict[src]
        else:
            return convert_dict



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import warnings

warnings.filterwarnings("ignore")
import sys

from ..utils import print_sys
from . import bi_pred_dataset, multi_pred_dataset
from ..metadata import dataset_names


class TestMultiPred(bi_pred_dataset.DataLoader):
    """Summary

    Attributes:
        entity1_name (str): Description
        entity2_name (str): Description
        two_types (bool): Description
    """

    def __init__(self, name, path="./data", label_name=None, print_stats=False):
        """Summary

        Args:
            name (TYPE): Description
            path (str, optional): Description
            label_name (None, optional): Description
            print_stats (bool, optional): Description
        """
        super().__init__(
            name,
            path,
            label_name,
            print_stats,
            dataset_names=dataset_names["test_multi_pred"],
        )
        self.entity1_name = "Antibody"
        self.entity2_name = "Antigen"
        self.two_types = True

        if print_stats:
            self.print_stats()

        print("Done!", flush=True, file=sys.stderr)



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT
"""
This file contains a primekg dataloader. 
"""

import numpy as np
import warnings

from ..utils import general_load
from ..utils.knowledge_graph import KnowledgeGraph

warnings.filterwarnings("ignore")


class PrimeKG(KnowledgeGraph):
    """PrimeKG data loader class to load the knowledge graph with additional support functions.
    """

    def __init__(self, path="./data"):
        """load the KG to the specified path"""
        df = general_load("primekg", path, ",")
        self.df = df
        self.path = path
        super().__init__(self.df)

    def get_data(self):
        return self.df

    def to_nx(self):
        import networkx as nx

        G = nx.Graph()
        for i in self.df.relation.unique():
            G.add_edges_from(
                self.df[self.df.relation == i][["x_name", "y_name"]].values,
                relation=i)
        return G

    def get_features(self, feature_type):
        if feature_type not in ["drug", "disease"]:
            raise ValueError("feature_type only supports drug/disease!")
        return general_load("primekg_" + feature_type + "_feature", self.path,
                            "\t")

    def get_node_list(self, node_type):
        df = self.df
        return np.unique(df[(df.x_type == node_type)].x_id.unique().tolist() +
                         df[(df.y_type == node_type)].y_id.unique().tolist())



from .adme import ADME
from .crispr_outcome import CRISPROutcome
from .develop import Develop
from .epitope import Epitope
from .hts import HTS
from .paratope import Paratope
from .qm import QM
from .test_single_pred import TestSinglePred
from .tox import Tox
from .yields import Yields



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import pandas as pd
import numpy as np
import os, sys, json
import warnings

warnings.filterwarnings("ignore")

from .base_group import BenchmarkGroup
from ..utils import bm_group_load, print_sys, fuzzy_search
from ..metadata import (
    get_task2category,
    bm_metric_names,
    benchmark_names,
    bm_split_names,
    docking_target_info,
)
from ..evaluator import Evaluator


class docking_group(BenchmarkGroup):
    """Create a docking group benchmark loader.

    Args:
        path (str, optional): the folder path to save/load the benchmarks.
        pyscreener_path (str, optional): the path to pyscreener repository in order to call docking scores.
        num_workers (int, optional): number of workers to parallelize dockings
        num_cpus (int, optional): number of CPUs assigned to docking
        num_max_call (int, optional): maximum number of oracle calls


    """

    def __init__(self,
                 path="./data",
                 num_workers=None,
                 num_cpus=None,
                 num_max_call=5000):
        """Create a docking group benchmark loader.

        Raises:
            ValueError: missing path to pyscreener.
        """
        super().__init__(name="Docking_Group", path=path, file_format="oracle")

        # if pyscreener_path is not None:
        # 	self.pyscreener_path = pyscreener_path
        # else:
        # 	raise ValueError("Please specify pyscreener_path!")

        if (num_workers is None) and (num_cpus is None):
            ## automatic selections
            cpu_total = os.cpu_count()
            if cpu_total > 1:
                num_cpus = 2
            else:
                num_cpus = 1
            num_workers = int(cpu_total / num_cpus)

        self.num_workers = num_workers
        self.num_cpus = num_cpus
        self.num_max_call = num_max_call
        from ..oracles import Oracle

    def __iter__(self):
        """iterate docking targets

        Returns:
            docking_group: the docking group class itself
        """
        self.index = 0
        self.num_datasets = len(self.dataset_names)
        return self

    def __next__(self):
        """retrieve the next benchmark

        Returns:
            dict: a dictionary of oracle function, molecule library dataset, and the name of docking target

        Raises:
            StopIteration: stop when all benchmarks are obtained.
        """
        if self.index < self.num_datasets:
            dataset = self.dataset_names[self.index]
            print_sys("--- " + dataset + " ---")

            data_path = os.path.join(self.path, dataset)
            if not os.path.exists(data_path):
                os.mkdir(data_path)

            target_pdb_file = os.path.join(self.path, dataset + ".pdb")
            self.index += 1

            from ..oracles import Oracle

            # oracle = Oracle(name = "Docking_Score", software="vina",
            # 	pyscreener_path = self.pyscreener_path,
            # 	receptors=[target_pdb_file],
            # 	center=docking_target_info[dataset]['center'], size=docking_target_info[dataset]['size'],
            # 	buffer=10, path=data_path, num_worker=self.num_workers, ncpu=self.num_cpus, num_max_call = self.num_max_call)
            oracle = Oracle(
                name="Docking_Score",
                pyscreener_path=self.pyscreener_path,
                receptors=[target_pdb_file],
                center=docking_target_info[dataset]["center"],
                size=docking_target_info[dataset]["size"],
                buffer=10,
                path=data_path,
                num_worker=self.num_workers,
                ncpu=self.num_cpus,
                num_max_call=self.num_max_call,
            )
            data = pd.read_csv(os.path.join(self.path, "zinc.tab"), sep="\t")
            return {"oracle": oracle, "data": data, "name": dataset}
        else:
            raise StopIteration

    def get_train_valid_split(self, seed, benchmark, split_type="default"):
        """no split for docking group

        Raises:
            ValueError: no split for docking group
        """
        raise ValueError(
            "Docking molecule generation does not have the concept of training/testing split! Checkout the usage in tdcommons.ai !"
        )

    def get(self, benchmark, num_max_call=5000):
        """retrieve one benchmark given benchmark name (docking target)

        Args:
            benchmark (str): the name of the benchmark
            num_max_call (int, optional): maximum of oracle calls

        Returns:
            dict: a dictionary of oracle function, molecule library dataset, and the name of docking target
        """
        dataset = fuzzy_search(benchmark, self.dataset_names)
        data_path = os.path.join(self.path, dataset)
        target_pdbqt_file = os.path.join(self.path, dataset + ".pdbqt")

        from ..oracles import Oracle

        # oracle = Oracle(name = "Docking_Score", software="vina",
        # 	pyscreener_path = self.pyscreener_path,
        # 	receptors=[target_pdb_file],
        # 	center=docking_target_info[dataset]['center'], size=docking_target_info[dataset]['size'],
        # 	buffer=10, path=data_path, num_worker=self.num_workers, ncpu=self.num_cpus, num_max_call = num_max_call)
        # oracle = Oracle(name = "Docking_Score",
        # 	receptor_pdbqt_file=target_pdbqt_file,
        # 	center=docking_target_info[dataset]['center'],
        # 	box_size=docking_target_info[dataset]['size'],
        # 	num_max_call = num_max_call)
        oracle = Oracle(name="3pbl_docking")
        data = pd.read_csv(os.path.join(self.path, "zinc.tab"), sep="\t")
        return {"oracle": oracle, "data": data, "name": dataset}

    def evaluate(self,
                 pred,
                 true=None,
                 benchmark=None,
                 m1_api=None,
                 save_dict=True):
        """Summary

        Args:
            pred (dict): a nested dictionary, where the first level key is the docking target, the value is another dictionary where the key is the maximum oracle calls, and value can have two options. One, a dictionary of SMILES paired up with the docking scores and Second, a list of SMILES strings, where the function will generate the docking scores automatically.
            benchmark (str, optional): name of the benchmark docking target.
            m1_api (str, optional): API token of Molecule.One. This is to use M1 service to generate synthesis score.
            save_dict (bool, optional): whether or not to save the results.

        Returns:
            dict: result with all realistic metrics generated

        Raises:
            ValueError: Description
        """
        results_all = {}

        for data_name, pred_all in pred.items():
            results_max_call = {}
            for num_max_call, pred_ in pred_all.items():

                results = {}

                recalc = False

                if isinstance(pred_, dict):
                    print_sys(
                        "The input is a dictionary, expected to have SMILES string as key and docking score as value!"
                    )
                    docking_scores = pred_
                    pred_ = list(pred_.keys())
                elif isinstance(pred_, list):
                    recalc = True
                    print_sys(
                        "The input is a list, docking score will be computed! If you already have the docking scores, please make the list as a dictionary with SMILES string as key and docking score as value"
                    )
                else:
                    raise ValueError(
                        "The input prediction must be a dictionary with SMILES and their docking scores or a list of SMILES!"
                    )
                ## pred is a list of smiles strings or a dictionary of smiles strings if docking scores are already calculated...
                if len(pred_) != 100:
                    raise ValueError(
                        "The expected output is a list/dictionary of top 100 molecules!"
                    )

                if recalc:
                    dataset = fuzzy_search(data_name, self.dataset_names)

                    # docking scores for the top K smiles (K <= 100)
                    target_pdb_file = os.path.join(self.path, dataset + ".pdb")
                    from ..oracles import Oracle

                    data_path = os.path.join(self.path, dataset)
                    oracle = Oracle(
                        name="Docking_Score",
                        software="vina",
                        pyscreener_path=self.pyscreener_path,
                        receptors=[target_pdb_file],
                        center=docking_target_info[dataset]["center"],
                        size=docking_target_info[dataset]["size"],
                        buffer=10,
                        path=data_path,
                        num_worker=self.num_workers,
                        ncpu=self.num_cpus,
                        num_max_call=10000,
                    )

                    docking_scores = oracle(pred_)
                print_sys("---- Calculating average docking scores ----")
                if len(
                        np.where(np.array(list(docking_scores.values())) > 0)
                    [0]) > 0.7:
                    ## check if the scores are all positive.. if so, make them all negative
                    docking_scores = {j: -k for j, k in docking_scores.items()}
                if save_dict:
                    results["docking_scores_dict"] = docking_scores
                values = np.array(list(docking_scores.values()))
                results["top100"] = np.mean(values)
                results["top10"] = np.mean(sorted(values)[:10])
                results["top1"] = min(values)

                # if m1_api is None:
                # 	print_sys('Ignoring M1 Synthesizability Evaluations. You can still submit your results without m1 score. Although for the submission, we encourage inclusion of m1 scores. To opt-in, set the m1_api to the token obtained via: https://tdcommons.ai/functions/oracles/#moleculeone')
                # else:
                # 	print_sys("---- Calculating molecule.one synthesizability score ----")
                # 	from ..oracles import Oracle
                # 	m1 = Oracle(name = 'Molecule One Synthesis', api_token = m1_api)
                # 	import heapq
                # 	from operator import itemgetter
                # 	top10_docking_smiles = list(dict(heapq.nsmallest(10, docking_scores.items(), key=itemgetter(1))).keys())
                # 	m1_scores = m1(top10_docking_smiles)
                # 	scores_array = list(m1_scores.values())
                # 	scores_array = np.array([float(i) for i in scores_array])
                # 	scores_array[np.where(scores_array == -1.0)[0]] = 10 # m1 score errors are usually large complex molecules
                # 	if save_dict:
                # 		results['m1_dict'] = m1_scores
                # 	results['m1'] = np.mean(scores_array)

                print_sys("---- Calculating synthetic accessibility score ----")
                from ..oracles import Oracle

                sa = Oracle(name="SA")
                scores_array = sa(pred_)
                if save_dict:
                    results["sa_dict"] = scores_array
                results["sa"] = np.mean(scores_array)

                print_sys("---- Calculating molecular filters scores ----")
                from ..chem_utils.oracle.filter import MolFilter

                ## follow guacamol
                filters = MolFilter(
                    filters=["PAINS", "SureChEMBL", "Glaxo"],
                    property_filters_flag=False,
                )
                pred_filter = filters(pred_)
                if save_dict:
                    results["pass_list"] = pred_filter
                results["%pass"] = float(len(pred_filter)) / 100
                results["top1_%pass"] = min(
                    [docking_scores[i] for i in pred_filter])
                print_sys("---- Calculating diversity ----")
                from ..evaluator import Evaluator

                evaluator = Evaluator(name="Diversity")
                score = evaluator(pred_)
                results["diversity"] = score
                print_sys("---- Calculating novelty ----")
                evaluator = Evaluator(name="Novelty")
                training = pd.read_csv(os.path.join(self.path, "zinc.tab"),
                                       sep="\t")
                score = evaluator(pred_, training.smiles.values)
                results["novelty"] = score
                results["top smiles"] = [
                    i[0]
                    for i in sorted(docking_scores.items(), key=lambda x: x[1])
                ]
                results_max_call[num_max_call] = results
            results_all[data_name] = results_max_call
        return results_all

    def evaluate_many(self,
                      preds,
                      save_file_name=None,
                      m1_api=None,
                      results_individual=None):
        """evaluate many runs together and output submission ready pkl file.

        Args:
            preds (list): a list of pred across runs, where each follows the format of pred in 'evaluate' function.
            save_file_name (str, optional): the name of the file to save the result.
            m1_api (str, optional): m1 API token for molecule synthesis score.
            results_individual (list, optional): if you already have generated the result from the evaluate function for each run, simply put in a list and it will not regenerate the results.

        Returns:
            dict: the output result file.
        """
        min_requirement = 3
        if len(preds) < min_requirement:
            return ValueError("Must have predictions from at least " +
                              str(min_requirement) +
                              " runs for leaderboard submission")
        if results_individual is None:
            individual_results = []
            for pred in preds:
                retval = self.evaluate(pred, m1_api=m1_api)
                individual_results.append(retval)
        else:
            individual_results = results_individual

        metrics = [
            "top100",
            "top10",
            "top1",
            "diversity",
            "novelty",
            "%pass",
            "top1_%pass",
            "sa",
            "top smiles",
        ]
        num_folds = len(preds)

        results_agg = {}

        for target in list(individual_results[0].keys()):
            results_agg_target = {}
            for num_calls in individual_results[0][target].keys():
                results_agg_target_call = {}
                for metric in metrics:
                    if metric == "top smiles":
                        results_agg_target_call[metric] = np.unique(
                            np.array([
                                individual_results[fold][target][num_calls]
                                [metric] for fold in range(num_folds)
                            ]).reshape(-1)).tolist()
                    else:
                        res = [
                            individual_results[fold][target][num_calls][metric]
                            for fold in range(num_folds)
                        ]
                        results_agg_target_call[metric] = [
                            round(np.mean(res), 3),
                            round(np.std(res), 3),
                        ]
                results_agg_target[num_calls] = results_agg_target_call
            results_agg[target] = results_agg_target

        import pickle

        if save_file_name is None:
            save_file_name = "tdc_docking_result"
        with open(save_file_name + ".pkl", "wb") as f:
            pickle.dump(results_agg, f)
        return results_agg



from .data_feature_generator import DataFeatureGenerator


class CellXGeneFeatureGenerator(DataFeatureGenerator):

    @classmethod
    def get_dense_soma_dataframe(cls, dataset):
        return dataset[dataset["soma_data"] != 0]

    @classmethod
    def format_cellxgene_dataframe(cls, dataset):
        dataset.columns = ["cell_idx", "gene_idx", "expression"]
        return dataset



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT"

import warnings

warnings.filterwarnings("ignore")

from . import generation_dataset
from ..metadata import dataset_names


class Reaction(generation_dataset.PairedDataLoader):
    """Data loader class accessing to forward reaction prediction task."""

    def __init__(
        self,
        name,
        path="./data",
        print_stats=False,
        input_name="reactant",
        output_name="product",
    ):
        """To create an data loader object for forward reaction prediction task. The goal is to predict
        the reaction products given a set of reactants

        Args:
            name (str): the name of the datset
            path (str, optional): the path to the saved data file.
            print_stats (bool, optional): whether to print the basic statistics
            input_name (str, optional): the name of the column containing input molecular data (reactant)
            output_name (str, optional): the name of the column containing output molecular data (product)
        """
        super().__init__(name, path, print_stats, input_name, output_name)



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import warnings

warnings.filterwarnings("ignore")

from . import generation_dataset
from ..metadata import dataset_names


class MolGen(generation_dataset.DataLoader):
    """Data loader class accessing to molecular generation task (distribution learning)"""

    def __init__(self,
                 name,
                 path="./data",
                 print_stats=False,
                 column_name="smiles"):
        """To create an data loader object for molecular generation task. The goal is to generate diverse,
        novel molecules that has desirable chemical properties. One can combined with oracle functions.

        Args:
            name (str): the name of the datset
            path (str, optional): the path to the saved data file.
            print_stats (bool, optional): whether to print the basic statistics
            column_name (str, optional): the name of the column containing molecular data.
        """
        super().__init__(name, path, print_stats, column_name)



from .config import DatasetConfig
from ..feature_generators.anndata_to_dataframe import AnnDataToDataFrame


class SCPerturb(DatasetConfig):
    """Configuration for the scPerturb datasets"""

    def __init__(self):
        super(SCPerturb, self).__init__(
            data_processing_class=AnnDataToDataFrame,
            functions_to_run=["anndata_to_df"],
            args_for_functions=[
                {
                    "obs_cols": [
                        "ncounts", 'celltype', 'cell_line', 'cancer', 'disease',
                        'tissue_type', 'perturbation', 'perturbation_type',
                        'ngenes'
                    ],
                },
            ],
        )


class SCPerturb_Gene(DatasetConfig):
    """Configuration for the scPerturb genetic perturbation datasets"""

    def __init__(self):
        super(SCPerturb_Gene, self).__init__(
            data_processing_class=AnnDataToDataFrame,
            functions_to_run=["anndata_to_df"],
            args_for_functions=[
                {
                    "obs_cols": [
                        'UMI_count', 'cancer', 'cell_line', 'disease',
                        'guide_id', 'ncounts', 'ngenes', 'nperts', 'organism',
                        'percent_mito', 'percent_ribo', 'perturbation',
                        'perturbation_type', 'tissue_type'
                    ],
                },
            ],
        )



"""A python module to build, handle, explore, and manipulate knowledge graphs.
"""

import pandas as pd
from copy import copy

kg_columns = [
    'relation', 'display_relation', 'x_id', 'x_type', 'x_name', 'x_source',
    'y_id', 'y_type', 'y_name', 'y_source'
]


class KnowledgeGraph:

    def __init__(self, df=None):
        if df is not None:
            self.df = df
        else:
            self.df = pd.DataFrame('', columns=kg_columns)

    def copy(self):
        return copy(self)

    def run_query(self, query, inplace=True):
        """build subgraph using given query"""
        df_filt = self.df.query(query).reset_index(drop=True)
        if inplace:
            self.df_raw = self.df
            self.df = df_filt
        else:
            return df_filt

    def get_nodes_by_source(self, source):
        # extract x nodes
        x_df = self.df.query(
            f"x_source == '{source}' | y_source == '{source}'")[[
                col for col in self.df.columns if col.startswith("x_")
            ]]

        for col in x_df.columns:
            x_df = x_df.rename(columns={col: col[2:]})

        # extract y nodes
        y_df = self.df.query(
            f"x_source == '{source}' | y_source == '{source}'")[[
                col for col in self.df.columns if col.startswith("y_")
            ]]
        for col in y_df.columns:
            y_df = y_df.rename(columns={col: col[2:]})
        # merge x and y nodes and keep only unique nodes
        out = pd.concat([
            x_df, y_df
        ], axis=0).query(f'source == "{source}"').drop_duplicates().reset_index(
            drop=True)

        return out


def build_KG(indices, relation, display_relation, x_id, x_type, x_name,
             x_source, y_id, y_type, y_name, y_source):
    df = pd.DataFrame('', columns=kg_columns, index=indices)

    df.relation = relation
    df.display_relation = display_relation

    df.x_id = x_id
    df.x_type = x_type
    df.x_name = x_name
    df.x_source = x_source

    df.y_id = y_id
    df.y_type = y_type
    df.y_name = y_name
    df.y_source = y_source

    kg = KnowledgeGraph(df)

    return kg



# Molecular Property Cliff Task
# https://github.com/bidd-group/MPCD

# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import pandas as pd
import sys
import warnings

warnings.filterwarnings("ignore")

from . import single_pred_dataset
from ..utils import create_fold_setting_cold, create_scaffold_split
from ..metadata import dataset_names


class MPC(single_pred_dataset.DataLoader):

    def __init__(self, name, path="./data"):
        self.name = name
        self.data = None
        self.is_molace = False

    def get_from_gh(self, link):
        import pandas as pd
        import requests
        import io

        data = requests.get(link)
        try:
            data.raise_for_status()
        except:
            raise Exception(
                "invalid link provided. choose a link for datasets in https://github.com/bidd-group/MPCD"
            )
        self.data = pd.read_csv(io.StringIO(data.text))
        return self.data

    def get_data(self, link=None, get_from_gh=True, **kwargs):
        link = link if link is not None else self.name
        if get_from_gh:
            return self.get_from_gh(link)
        # support direct interfface with MoleculeACE API as well
        from MoleculeACE import Data, Descriptors
        self.molace = True
        try:
            self.data = Data(self.name)
            self.data(Descriptors.SMILES)
        except:
            raise Exception(
                "could not find dataset {}. For list of MoleculeAce datasets see https://github.com/bidd-group/MPCD/tree/main?tab=readme-ov-file#overview-of-the-mpc-benchmark-datasets"
                .format(self.name))
        return self.data

    def get_split(self, method="scaffold", seed=42, frac=[0.7, 0.1, 0.2]):
        d = self.get_data()
        if not self.is_molace:
            if method == "scaffold":
                return create_scaffold_split(d,
                                             seed=seed,
                                             frac=frac,
                                             entity="SMILES")
            elif method == "cold":
                return create_fold_setting_cold(d,
                                                seed=seed,
                                                frac=frac,
                                                entities="SMILES")
            raise Exception(
                "only scaffold or cold splits supported for the MPC task")
        train = pd.concat([d.x_train, d.y_train], axis=1)
        test = pd.concat([d.x_test, d.y_test], axis=1)
        return {
            "train": train,
            "test": test,
        }



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import warnings

warnings.filterwarnings("ignore")
import sys

from ..utils import print_sys
from . import bi_pred_dataset, multi_pred_dataset
from ..metadata import dataset_names


class TrialOutcome(multi_pred_dataset.DataLoader):
    """Data loader class to load datasets in clinical trial outcome Prediction task.
    More info: https://tdcommons.ai/multi_pred_tasks/trialoutcome/


    Task Description: Binary Classification.
                      Given the drug molecule, disease code (ICD) and trial protocol (eligibility criteria),
                      predict their trial approval rate.

    Args:
        name (str): the dataset name.
        path (str, optional):
            The path to save the data file, defaults to './data'
        label_name (str, optional):
            For multi-label dataset, specify the label name, defaults to None
        print_stats (bool, optional):
            Whether to print basic statistics of the dataset, defaults to False

    """

    def __init__(self, name, path="./data", label_name=None, print_stats=False):
        """Create Clinical Trial Outcome Prediction dataloader object"""
        super().__init__(name,
                         path,
                         print_stats,
                         dataset_names=dataset_names["TrialOutcome"])
        self.entity1_name = "drug_molecule"
        self.entity2_name = "disease_code"
        # self.entity3_name = "eligibility_criteria"

        if print_stats:
            self.print_stats()

        print("Done!", flush=True, file=sys.stderr)



from setuptools import find_packages, setup

# read the contents of README file
from os import path
from io import open  # for Python 2 and 3 compatibility

# get __version__ from _version.py
ver_file = path.join("tdc", "version.py")
with open(ver_file) as f:
    exec(f.read())

this_directory = path.abspath(path.dirname(__file__))


# read the contents of README.md
def readme():
    with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
        return f.read()


# read the contents of requirements.txt
with open(path.join(this_directory, "requirements.txt"),
          encoding="utf-8") as f:
    requirements = f.read().splitlines()

setup(
    name="PyTDC",
    version=__version__,
    license="MIT",
    description="Therapeutics Data Commons",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/mims-harvard/TDC",
    author="TDC Team",
    author_email="alejandro_velez-arce@hms.harvard.edu",
    packages=find_packages(exclude=["test"]),
    zip_safe=False,
    include_package_data=True,
    install_requires=requirements,
    setup_requires=["setuptools>=38.6.0"],
)



"""miscellaneous utilities functions
"""
import os, sys
import numpy as np
import pandas as pd
import subprocess
import pickle
from fuzzywuzzy import fuzz


def fuzzy_search(name, dataset_names):
    """fuzzy matching between the real dataset name and the input name

    Args:
        name (str): input dataset name given by users
        dataset_names (str): the exact dataset name in TDC

    Returns:
        s: the real dataset name

    Raises:
        ValueError: the wrong task name, no name is matched
    """
    name = name.lower()
    if name[:4] == "tdc.":
        name = name[4:]
    if name in dataset_names:
        s = name
    else:
        # print("========fuzzysearch=======", dataset_names, name)
        s = get_closet_match(dataset_names, name)[0]
    if s in dataset_names:
        return s
    else:
        raise ValueError(
            s +
            " does not belong to this task, please refer to the correct task name!"
        )


def get_closet_match(predefined_tokens, test_token, threshold=0.8):
    """Get the closest match by Levenshtein Distance.

    Args:
        predefined_tokens (list): Predefined string tokens.
        test_token (str): User input that needs matching to existing tokens.
        threshold (float, optional): The lowest match score to raise errors, defaults to 0.8

    Returns:
        str: the exact token with highest matching prob
            float: probability

    Raises:
        ValueError: no name is matched
    """
    prob_list = []

    for token in predefined_tokens:
        # print(token)
        prob_list.append(fuzz.ratio(
            str(token).lower(),
            str(test_token).lower()))

    assert len(prob_list) == len(predefined_tokens)

    prob_max = np.nanmax(prob_list)
    token_max = predefined_tokens[np.nanargmax(prob_list)]

    # match similarity is low
    if prob_max / 100 < threshold:
        print_sys(predefined_tokens)
        raise ValueError(
            test_token, "does not match to available values. "
            "Please double check.")
    return token_max, prob_max / 100


def save_dict(path, obj):
    """save an object to a pickle file

    Args:
        path (str): the path to save the pickle file
        obj (object): any file
    """
    with open(path, "wb") as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)


def load_dict(path):
    """load an object from a path

    Args:
        path (str): the path where the pickle file locates

    Returns:
        object: loaded pickle file
    """
    with open(path, "rb") as f:
        return pickle.load(f)


def install(package):
    """install pip package

    Args:
        package (str): package name
    """
    subprocess.check_call([sys.executable, "-m", "pip", "install", package])


def print_sys(s):
    """system print

    Args:
        s (str): the string to print
    """
    print(s, flush=True, file=sys.stderr)


def to_submission_format(results):
    """convert the results to submission-ready format in leaderboard

    Args:
        results (dict): a dictionary of metrics across five runs

    Returns:
        dict: a dictionary of metrics and values with mean and std
    """
    df = pd.DataFrame(results)

    def get_metric(x):
        metric = []
        for i in x:
            metric.append(list(i.values())[0])
        return [round(np.mean(metric), 3), round(np.std(metric), 3)]

    return dict(df.apply(get_metric, axis=1))



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import pandas as pd
import numpy as np
import os, sys, json
import warnings

warnings.filterwarnings("ignore")

from .. import base_dataset
from ..utils import (
    distribution_dataset_load,
    generation_paired_dataset_load,
    three_dim_dataset_load,
    print_sys,
)
from ..utils import create_fold


class DataLoader(base_dataset.DataLoader):
    """A base dataset loader class.

    Attributes:
        dataset_names (str): name of the dataset.
        name (str): The name fo the dataset.
        path (str): the path to save the data file.
        smiles_lst (list): a list of smiles strings as training data for distribution learning.
    """

    def __init__(self, name, path, print_stats, column_name):
        """To create a base dataloader object that each generation task can inherit from.

        Args:
            name (str): the name of the dataset.
            path (str): the path to save the data file.
            print_stats (bool): whether to print the basic statistics of the dataset.
            column_name (str): The name of the column containing smiles strings.
        """
        from ..metadata import single_molecule_dataset_names

        self.smiles_lst = distribution_dataset_load(
            name, path, single_molecule_dataset_names, column_name=column_name)
        ### including fuzzy-search
        self.name = name
        self.path = path
        self.dataset_names = single_molecule_dataset_names
        if print_stats:
            self.print_stats()
        print_sys("Done!")

    def print_stats(self):
        """Print the basic statistics of the dataset."""
        print(
            "There are " + str(len(self.smiles_lst)) + " molecules ",
            flush=True,
            file=sys.stderr,
        )

    def get_data(self, format="df"):
        """Return the data from the whole dataset.

        Args:
            format (str, optional): the desired format for molecular data.

        Returns:
            pandas DataFrame/dict: a dataframe of the dataset/a distionary for information

        Raises:
            AttributeError: Use the correct format as input (df, dict)
        """
        if format == "df":
            return pd.DataFrame({"smiles": self.smiles_lst})
        elif format == "dict":
            return {"smiles": self.smiles_lst}
        else:
            raise AttributeError("Please use the correct format input")

    def get_split(self, method="random", seed=42, frac=[0.7, 0.1, 0.2]):
        """Return the data splitted as train, valid, test sets.

        Arguments:
            method (str): splitting schemes: random, scaffold
            seed (int): random seed, default 42
            frac (list of float): ratio of train/val/test split

        Returns:
            pandas DataFrame/dict: a dataframe of the dataset

        Raises:
            AttributeError: Use the correct split method as input (random, scaffold)
        """
        df = self.get_data(format="df")

        if method == "random":
            return create_fold(df, seed, frac)
        else:
            raise AttributeError("Please use the correct split method")


class PairedDataLoader(base_dataset.DataLoader):
    """A basic class for generation of biomedical entities conditioned on other entities, such as reaction prediction.

    Attributes:
        dataset_names (str): the name fo the dataset.
        name (str): the name of the dataset.
        path (str): the path to save the data file.
    """

    def __init__(self, name, path, print_stats, input_name, output_name):
        """To create a object for paired biomedical entities generation.

        Arguments:
            name (str): fuzzy name of the generation dataset. e.g., uspto50k, qed, drd, ...
            path (str): directory path that stores the dataset, e.g., ./data
            print_stats (bool): whether print the stats.
            input_name (str): The column name of input biomedical entities.
            output_name (str): The column name of output biomedical entities.

        """
        from ..metadata import paired_dataset_names

        self.input_smiles_lst, self.output_smiles_lst = generation_paired_dataset_load(
            name, path, paired_dataset_names, input_name,
            output_name)  ### including fuzzy-search
        self.name = name
        self.path = path
        self.dataset_names = paired_dataset_names
        if print_stats:
            self.print_stats()
        print_sys("Done!")

    def print_stats(self):
        """Print the statistics of the dataset."""
        print(
            "There are " + str(len(self.input_smiles_lst)) + " paired samples",
            flush=True,
            file=sys.stderr,
        )

    def get_data(self, format="df"):
        """Return the data from the whole dataset.

        Args:
            format (str, optional): the desired format for molecular data.

        Returns:
            pandas DataFrame/dict: a dataframe of the dataset/a distionary for information

        Raises:
            AttributeError: Use the correct format as input (df, dict)
        """
        if format == "df":
            return pd.DataFrame({
                "input": self.input_smiles_lst,
                "output": self.output_smiles_lst
            })
        elif format == "dict":
            return {
                "input": self.input_smiles_lst,
                "output": self.output_smiles_lst
            }
        else:
            raise AttributeError("Please use the correct format input")

    def get_split(self, method="random", seed=42, frac=[0.7, 0.1, 0.2]):
        """Return the data splitted as train, valid, test sets.

        Arguments:
            method (str): splitting schemes: random, scaffold
            seed (int): random seed, default 42
            frac (list of float): ratio of train/val/test split

        Returns:
            pandas DataFrame/dict: a dataframe of the dataset

        Raises:
            AttributeError: Use the correct split method as input (random, scaffold)
        """

        df = self.get_data(format="df")

        if method == "random":
            return create_fold(df, seed, frac)
        else:
            raise AttributeError("Please use the correct split method")


class DataLoader3D(base_dataset.DataLoader):
    """A basic class for generation of 3D biomedical entities. (under construction)

    Attributes:
        df (str): the dataset in pandas DataFrame format.
        name (str): the name of the dataset.
        path (str): the path to save the data file.
    """

    ### locally, unzip a folder, with the main file the dataframe with SMILES, Mol Object for various kinds of entities.
    ### also, for each column, contains a sdf file.

    def __init__(self, name, path, print_stats, dataset_names, column_name):
        """To create an object for 3D biomedical entities generation.

        Args:
            name (str): the name of the dataset.
            path (str): the path to save the data file.
            print_stats (bool): whether to print the basic statistics of the dataset.
            column_name (str): The name of the column containing smiles strings.
        """
        self.df, self.path, self.name = three_dim_dataset_load(
            name, path, dataset_names)
        if print_stats:
            self.print_stats()
        print_sys("Done!")

    def print_stats(self):
        """Print the basic statistics of the dataset."""
        print(
            "There are " + str(len(self.df)) + " data points ",
            flush=True,
            file=sys.stderr,
        )

    def get_data(self, format="df", more_features="None"):
        """Return the data from the whole dataset.

        Args:
            format (str, optional): the desired format for molecular data.
            more_features (str, optional): 3D feature format, choose from [Graph3D, Coulumb]

        Returns:
            pandas DataFrame/dict: a dataframe of the dataset/a distionary for information

        Raises:
            AttributeError: Use the correct format as input (df, dict)
            ImportError: Please install rdkit by 'conda install -c conda-forge rdkit'

        """
        if more_features in ["None", "SMILES"]:
            pass
        elif more_features in ["Graph3D", "Coulumb",
                               "SELFIES"]:  # why SELFIES here?
            try:
                from rdkit.Chem.PandasTools import LoadSDF
                from rdkit import rdBase

                rdBase.DisableLog("rdApp.error")
            except:
                raise ImportError(
                    "Please install rdkit by 'conda install -c conda-forge rdkit'! "
                )

            from ..chem_utils import MolConvert
            from ..metadata import sdf_file_names

            convert = MolConvert(src="SDF", dst=more_features)
            for i in sdf_file_names[self.name]:
                self.df[i + "_" + more_features] = convert(self.path + i +
                                                           ".sdf")

        if format == "df":
            return self.df
        else:
            raise AttributeError("Please use the correct format input")

    def get_split(self, method="random", seed=42, frac=[0.7, 0.1, 0.2]):
        """Return the data splitted as train, valid, test sets.

        Arguments:
            method (str): splitting schemes: random, scaffold
            seed (int): random seed, default 42
            frac (list of float): ratio of train/val/test split

        Returns:
            pandas DataFrame/dict: a dataframe of the dataset

        Raises:
            AttributeError: Use the correct split method as input (random, scaffold)
        """
        df = self.get_data(format="df")

        if method == "random":
            return create_fold(df, seed, frac)
        else:
            raise AttributeError("Please use the correct split method")



"""
Class encapsulating general data processing functions. Also supports running them in sequence.
Goal is to make it easier to integrate custom datasets not yet in TDC format.
"""
import inspect
from pandas import DataFrame, isnull

from .base import FeatureGenerator


class DataFeatureGenerator(FeatureGenerator):
    """
    Class encapsulating general data processing functions. Also supports running them in sequence.
    Goals are to make it easier to integrate custom datasets not yet in TDC format.
    """

    def __init__(self):
        pass

    @classmethod
    def autofill_identifier(cls, dataset, autofill_column, key_column):
        """Autofill a column based on base column. Assumes one-to-one mapping between both.
        Modifications done in-place.
        
        Args:
            dataset (pandas.DataFrame): dataset to modify.
            autofill_column (str): name of the column to autofill.
            key_column (str): name of the column used for indexing.
            
        Returns:
            pandas.DataFrame: The modified dataset.
        """
        # Create a mapping from key_column to autofill_column
        mapping = dataset.dropna(subset=[autofill_column]).drop_duplicates(
            subset=[key_column]).set_index(
                key_column)[autofill_column].to_dict()

        # Apply the mapping to fill missing values in autofill_column based on key_column values
        dataset[autofill_column] = dataset[key_column].map(mapping)

        return dataset

    @classmethod
    def create_range(cls, dataset, column, keys=None, subs=None):
        """From a column with numeric +/- values, create upper,lower, and expected columns
        Modifies dataset in-place.
        If special keys are provided, corresponding entries are replaced for the numerical value in subs"""

        def helper(entry):
            if isnull(entry):
                return (None, None)
            buffer = ""
            assert isinstance(entry, str), ("Entry must be a string", entry)
            for idx, char in enumerate(entry):
                if char.isdigit() or char == ".":
                    buffer += char
                else:
                    break
            rest = entry[idx + 1:]
            return float(buffer), float(rest)

        keys = [] if keys is None else keys
        subs = [] if subs is None else subs
        assert isinstance(keys, list)
        assert isinstance(subs, list)
        assert len(keys) == len(subs)
        subs_dict = {k: s for k, s in zip(keys, subs)}
        if None not in subs_dict:
            subs_dict[None] = None
        entries = [
            helper(x) if x not in subs_dict else (subs_dict[x], subs_dict[x])
            for x in dataset[column]
        ]
        bounds = [[x1 - x2, x1, x1 +
                   x2] if x1 not in subs_dict else [x1, x1, x1]
                  for x1, x2 in entries]
        df_bounds = DataFrame(bounds, columns=['lower', 'expected', 'upper'])
        dataset["lower"] = df_bounds["lower"]
        dataset["expected"] = df_bounds["expected"]
        dataset["upper"] = df_bounds["upper"]
        return dataset

    @classmethod
    def process_data(cls, dataset, functions, args):
        """Run a series of data transformation on the input dataset

        Args:
            dataset (no specificaton): The original dataset. As different functions can require different data types, this is left unspecified
            functions (list[str]): list of string with function names
            args (list[dict(str: object)]): a list of dicts containing the arguments for each function.
        """
        functions_dict = {
            name: getattr(cls, name) for name, _ in inspect.getmembers(cls)
        }
        df = dataset
        for idx, f in enumerate(functions):
            a = args[idx]
            a["dataset"] = df
            try:
                df = functions_dict[f](**a)
            except KeyError as e:
                raise ValueError("Unknown function '{}'".format(f),
                                 functions_dict.keys(), e)
        return df



from .config import DatasetConfig
from ..feature_generators.cellxgene_generator import CellXGeneFeatureGenerator


class CellXGeneConfig(DatasetConfig):

    def __init__(self):
        super(CellXGeneConfig, self).__init__(
            data_processing_class=CellXGeneFeatureGenerator,
            functions_to_run=[
                "get_dense_soma_dataframe", "format_cellxgene_dataframe"
            ],
            args_for_functions=[{}, {}, {}],
        )



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT"

import pandas as pd
import numpy as np
import os, sys, json
import warnings

warnings.filterwarnings("ignore")

from .. import base_dataset
from ..utils import bi_distribution_dataset_load, print_sys
from ..utils import create_combination_generation_split
from ..metadata import dataset_names


class SBDD(base_dataset.DataLoader):
    """Data loader class accessing to structure-based drug design task."""

    def __init__(
        self,
        name,
        path="./data",
        print_stats=False,
        return_pocket=False,
        threshold=15,
        remove_protein_Hs=True,
        remove_ligand_Hs=True,
        keep_het=False,
        save=True,
    ):
        """To create a base dataloader object for structure-based drug design task.

        Args:
            name (str): the name of the dataset.
            path (str): the path to save the data file.
            print_stats (bool): whether to print the basic statistics of the dataset.
            return_pocket (bool): whether to return only protein pocket or full protein.
                threshold (int): only enabled when return_pocket is to True, if pockets are not provided in the raw data,
                                                 the threshold is used as a radius for a sphere around the ligand center to consider protein pocket.
                remove_protein_Hs (bool): whether to remove H atoms from proteins or not.
                remove_ligand_Hs (bool): whether to remove H atoms from ligands or not.
                keep_het (bool): whether to keep het atoms (e.g. cofactors) in protein.
                allowed_atom_list (list(str)): atom types allowed to include.
                save (bool): whether to save preprocessed data and splits.
        """
        from ..metadata import multiple_molecule_dataset_names

        try:
            import biopandas
        except:
            raise ImportError(
                "Please install biopandas by 'pip install biopandas'! ")
        protein, ligand = bi_distribution_dataset_load(
            name,
            path,
            multiple_molecule_dataset_names,
            return_pocket,
            threshold,
            remove_protein_Hs,
            remove_ligand_Hs,
            keep_het,
        )
        if save:
            np.savez(
                os.path.join(path, name + ".npz"),
                protein_coord=protein["coord"],
                protein_atom=protein["atom_type"],
                ligand_coord=ligand["coord"],
                ligand_atom=ligand["atom_type"],
            )
        self.save = save

        self.ligand = ligand
        self.protein = protein

        ### including fuzzy-search
        self.name = name
        self.path = path
        self.dataset_names = multiple_molecule_dataset_names
        self.return_pocket = return_pocket
        if print_stats:
            self.print_stats()
        print_sys("Done!")

    def print_stats(self):
        """Print the basic statistics of the dataset."""
        print(
            "There are " + str(len(self.smiles_lst)) + " molecules ",
            flush=True,
            file=sys.stderr,
        )

    def get_data(self, format="dict"):
        """Return the data from the whole dataset.

        Args:
            format (str, optional): the desired format for molecular data.

        Returns:
            dict: a dict of protein, ligand information

        Raises:
            AttributeError: Use the correct format as input (dict)
        """
        if format == "dict":
            return {"protein": self.protein, "ligand": self.ligand}
        else:
            raise AttributeError("Please use the correct format input")

    def get_split(self, method="random", seed=42, frac=[0.7, 0.1, 0.2]):
        """Return the data splitted as train, valid, test sets.

        Arguments:
            method (str): splitting schemes: random, scaffold
            seed (int): random seed, default 42
            frac (list of float): ratio of train/val/test split

        Returns:
            pandas DataFrame/dict: a dataframe of the dataset

        Raises:
            AttributeError: Use the correct split method as input (random, scaffold)
        """
        data = self.get_data(format="dict")
        protein, ligand = data["protein"], data["ligand"]

        splitted_data = create_combination_generation_split(
            protein, ligand, seed, frac)

        if self.save:
            np.savez(
                os.path.join(self.path, self.name + "_train.npz"),
                protein_coord=splitted_data["train"]["protein_coord"],
                protein_atom=splitted_data["train"]["protein_atom_type"],
                ligand_coord=splitted_data["train"]["ligand_coord"],
                ligand_atom=splitted_data["train"]["ligand_atom_type"],
            )
            np.savez(
                os.path.join(self.path, self.name + "_valid.npz"),
                protein_coord=splitted_data["valid"]["protein_coord"],
                protein_atom=splitted_data["valid"]["protein_atom_type"],
                ligand_coord=splitted_data["valid"]["ligand_coord"],
                ligand_atom=splitted_data["valid"]["ligand_atom_type"],
            )
            np.savez(
                os.path.join(self.path, self.name + "_test.npz"),
                protein_coord=splitted_data["test"]["protein_coord"],
                protein_atom=splitted_data["test"]["protein_atom_type"],
                ligand_coord=splitted_data["test"]["ligand_coord"],
                ligand_atom=splitted_data["test"]["ligand_atom_type"],
            )

        if method == "random":
            return splitted_data
        else:
            raise AttributeError("Please use the correct split method")



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import warnings

warnings.filterwarnings("ignore")
import sys

from ..utils import print_sys
from . import bi_pred_dataset, multi_pred_dataset
from ..metadata import dataset_names


class DrugSyn(multi_pred_dataset.DataLoader):
    """Data loader class to load datasets in Drug Synergy Prediction task.
    More info: https://tdcommons.ai/multi_pred_tasks/drugsyn/

    Task Description: Regression.
                      Given the gene expression of cell lines and two SMILES strings of the drug combos,
                      predict the drug synergy level.

    Args:
        name (str): the dataset name.
        path (str, optional):
            The path to save the data file, defaults to './data'
        print_stats (bool, optional):
            Whether to print basic statistics of the dataset, defaults to False

    """

    def __init__(self, name, path="./data", print_stats=False):
        """Create Drug Synergy Prediction dataloader object"""
        super().__init__(name,
                         path,
                         print_stats,
                         dataset_names=dataset_names["DrugSyn"])

        if print_stats:
            self.print_stats()

        print("Done!", flush=True, file=sys.stderr)



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import pandas as pd
import numpy as np
import os, sys, json
import warnings

warnings.filterwarnings("ignore")

from .. import base_dataset
from ..utils import (
    dataset2target_lists,
    multi_dataset_load,
    create_fold,
    create_fold_setting_cold,
    create_combination_split,
    print_sys,
)


class DataLoader(base_dataset.DataLoader):
    """A base data loader class that each multi-instance prediction task dataloader class can inherit from.

    Attributes: TODO

    """

    def __init__(self, name, path, print_stats, dataset_names):
        """create dataloader object

        Args:
                name (str): name of dataloader
                path (str): the path where data is saved
                label_name (str): name of label
                print_stats (bool): whether to print statistics of dataset
                dataset_names (str): A list of dataset names available for a task
        """
        if name.lower() in dataset2target_lists.keys():
            if label_name is None:
                raise ValueError(
                    "Please select a label name. You can use tdc.utils.retrieve_label_name_list('"
                    + name.lower() +
                    "') to retrieve all available label names.")

        df = multi_dataset_load(name, path, dataset_names)

        self.df = df
        self.name = name
        self.path = path

    def get_data(self, format="df"):
        """generate data in some format, e.g., pandas.DataFrame

        Args:
            format (str, optional):
                format of data, the default value is 'df' (DataFrame)

        Returns:
            pandas DataFrame/dict: a dataframe of a dataset/a dictionary for key information in the dataset

        Raises:
            AttributeError: Use the correct format input (df, dict, DeepPurpose)
        """
        if format == "df":
            return self.df
        elif format == "dict":
            return dict(self.df)
        else:
            raise AttributeError("Please use the correct format input")

    def print_stats(self):
        """print the statistics of the dataset"""
        print_sys("--- Dataset Statistics ---")
        print(str(len(self.df)) + " data points.", flush=True, file=sys.stderr)
        print_sys("--------------------------")

    def get_split(self,
                  method="random",
                  seed=42,
                  frac=[0.7, 0.1, 0.2],
                  column_name=None,
                  data_ready=False):
        """split dataset into train/validation/test.

        Args:
            method (str, optional):
                split method, the default value is 'random'
            seed (int, optional):
                random seed, defaults to '42'
            frac (list, optional):
                train/val/test split fractions, defaults to '[0.7, 0.1, 0.2]'
            column_name (None, optional): Description

        Returns:
            dict: a dictionary with three keys ('train', 'valid', 'test'), each value is a pandas dataframe object of the splitted dataset

        Raises:
            AttributeError: the input split method is not available.

        """
        df = self.get_data(format="df") if not data_ready else self.df

        if method == "random":
            return create_fold(df, seed, frac)

        elif method == "cold_split":
            if isinstance(column_name, str):
                column_name = [column_name]
            if (column_name is None) or (not all(
                [x in df.columns.values for x in column_name])):
                raise AttributeError(
                    "For cold_split, please provide one or multiple column names that are contained in the dataframe."
                )
            return create_fold_setting_cold(df, seed, frac, column_name)

        elif method == "combination":
            return create_combination_split(df, seed, frac)
        else:
            raise AttributeError(
                "Please select a splitting strategy from random, cold_split, or combination."
            )



from collections.abc import Iterable
import pandas as pd
import tiledbsoma

from .anndata_dataset import DataLoader
from ..dataset_configs.cellxgene_config import CellXGeneConfig
from ..metadata import dataset_names
from ..resource.cellxgene_census import CensusResource


class CellXGeneTemplate(DataLoader):

    def __init__(self, name, path, print_stats, dsn=None):
        dsn = dataset_names["CellXGene"] if dsn is None else dsn
        super().__init__(name, path, print_stats, dsn)


class CellXGene(
        DataLoader
):  #TODO: create separate class for wrapping around a resource. as this is the only current use case it is left as is.

    def __init__(self, name, path="./data", print_stats=False):
        self.available_datasets = None
        self.resource = CensusResource(
        )  # only humans will be supported at this time
        self.is_collection = None
        self.genes = None
        self.config = CellXGeneConfig()
        self.data_processor = self.config.processing_callback
        if type(name) != list:
            if name not in self.get_available_datasets():
                raise ValueError("Invalid dataset name", name)
            self.is_collection = False
        elif not all(n in self.get_available_datasets() for n in name):
            raise Exception(
                "one or more input datasets in your list isn't in cellxgene available datasets",
                name)
        else:
            self.is_collection = True
        self.name = name
        assert self.is_collection is not None

    def get_data_for_dataset(self,
                             measurement_name=None,
                             value_filter=None,
                             value_adjustment=None):
        """Generator Expression"""
        measurement_name = measurement_name or "RNA"
        value_adjustment = value_adjustment or "raw"
        if self.genes is None:
            feature_presence_vector = self.resource.get_feature_dataset_presence_matrix_entry(
                self.name, measurement_name=measurement_name)
            genes_in_dataset = feature_presence_vector.values[0].tolist()
            self.genes = genes_in_dataset
        gene_filter = [idx for idx, v in enumerate(self.genes) if v]
        for slice in self.resource.query_measurement_matrix(
                value_filter=value_filter,
                measurement_name=measurement_name,
                value_adjustment=value_adjustment,
                gene_filter=gene_filter):
            filtered_df = self.data_processor(slice)
            yield filtered_df

    def get_data_for_collection(self,
                                measurement_name=None,
                                value_filter=None,
                                value_adjustment=None):
        """Generator Expression"""
        assert type(self.name) == list
        all_datasets = self.name
        self.genes = self.get_union_of_genes(measurement_name=measurement_name,
                                             all_datasets=all_datasets)
        yield from self.get_data_for_dataset(measurement_name=measurement_name,
                                             value_filter=value_filter,
                                             value_adjustment=value_adjustment)

    def get_union_of_genes(self, measurement_name=None, all_datasets=None):
        assert all_datasets is not None
        assert self.is_collection
        assert isinstance(all_datasets, Iterable)
        measurement_name = measurement_name or "RNA"
        gene_presence = None
        for ds in all_datasets:
            feature_presence_vector = self.resource.get_feature_dataset_presence_matrix_entry(
                ds, measurement_name=measurement_name)
            genes_in_dataset = feature_presence_vector.values[0].tolist()
            gene_presence = genes_in_dataset if gene_presence is None else [
                a | b for a, b in zip(genes_in_dataset, gene_presence)
            ]
        return gene_presence

    def get_data(self,
                 measurement_name=None,
                 value_filter=None,
                 value_adjustment=None):
        """TDC get_data() API implementation for the CELLXGENE data class. Can be a generator expression or return a traditional
        Pandas dataframe of union of all CELLXGENE entries of the queried dataset satisfying the provided conditions.

        Args:
            measurement_name (str, optional): measurement name to query, i.e. "RNA". Defaults to None.
            value_filter (str, optional): a valuer filter (obs) to apply to the query. Defaults to None.
            value_adjustment (str, optional): the type of count to obtain from count matricx for this measurement. Defaults to None.
            as_dataframe (bool, optional): Whether to return a fully concatendated dataframe or yield from the dataset. Defaults to False.

        Returns:
            pd.DataFrame: a dataframe if as_dataframe is True

        Yields:
            pd.DataFrame: yields chunks of the output datframe if as_dataframe is False

        WARNNG: requesting as_dataframe = True for too large a result may result in OOM or other process failure.
        """
        generator = self.get_data_for_collection(measurement_name=measurement_name, value_filter=value_filter, value_adjustment=value_adjustment) \
            if self.is_collection \
            else self.get_data_for_dataset(measurement_name=measurement_name, value_filter=value_filter, value_adjustment=value_adjustment)
        yield from generator

    def get_dataframe(self,
                      measurement_name=None,
                      value_filter=None,
                      value_adjustment=None,
                      debug=False):
        return pd.concat(self.get_data(measurement_name=measurement_name,
                                       value_filter=value_filter,
                                       value_adjustment=value_adjustment,
                                       debug=debug),
                         axis=0)

    def get_split(self,
                  measurement_name=None,
                  value_filter=None,
                  value_adjustment=None,
                  debug=False,
                  **kwargs):
        """TDC get_split() API implementation for the CELLXGENE data class. It assumes the dataset can fit in memory.
        Invokes the parent class get_split() function and supports configuring parameters for said class.

        WARNING: calling get_split() on too large a dataset can result in OOM or other process failure.
        """
        self.df = self.get_dataframe(measurement_name=measurement_name,
                                     value_filter=value_filter,
                                     value_adjustment=value_adjustment,
                                     debug=debug)
        assert len(self.df) > 1
        kwargs["data_ready"] = True
        return super().get_split(**kwargs)

    def get_available_datasets(self):
        if self.available_datasets is not None:
            return self.available_datasets
        self.available_datasets = self.resource.get_dataset_metadata()
        self.available_datasets = self.available_datasets[
            "dataset_title"].unique()
        return self.available_datasets



import numpy as np
from collections import defaultdict
from typing import List

try:
    from rdkit import Chem
    from rdkit import rdBase

    rdBase.DisableLog("rdApp.error")
except:
    raise ImportError(
        "Please install rdkit by 'conda install -c conda-forge rdkit'! ")

try:
    import networkx as nx
except:
    raise ImportError("Please install networkx by 'pip install networkx'! ")

from ...utils import print_sys

############## begin xyz2mol ################
# from https://github.com/jensengroup/xyz2mol/blob/master/xyz2mol.py

global __ATOM_LIST__
__ATOM_LIST__ = [
    "h",
    "he",
    "li",
    "be",
    "b",
    "c",
    "n",
    "o",
    "f",
    "ne",
    "na",
    "mg",
    "al",
    "si",
    "p",
    "s",
    "cl",
    "ar",
    "k",
    "ca",
    "sc",
    "ti",
    "v ",
    "cr",
    "mn",
    "fe",
    "co",
    "ni",
    "cu",
    "zn",
    "ga",
    "ge",
    "as",
    "se",
    "br",
    "kr",
    "rb",
    "sr",
    "y",
    "zr",
    "nb",
    "mo",
    "tc",
    "ru",
    "rh",
    "pd",
    "ag",
    "cd",
    "in",
    "sn",
    "sb",
    "te",
    "i",
    "xe",
    "cs",
    "ba",
    "la",
    "ce",
    "pr",
    "nd",
    "pm",
    "sm",
    "eu",
    "gd",
    "tb",
    "dy",
    "ho",
    "er",
    "tm",
    "yb",
    "lu",
    "hf",
    "ta",
    "w",
    "re",
    "os",
    "ir",
    "pt",
    "au",
    "hg",
    "tl",
    "pb",
    "bi",
    "po",
    "at",
    "rn",
    "fr",
    "ra",
    "ac",
    "th",
    "pa",
    "u",
    "np",
    "pu",
]

global atomic_valence
global atomic_valence_electrons

atomic_valence = defaultdict(list)
atomic_valence[1] = [1]
atomic_valence[5] = [3, 4]
atomic_valence[6] = [4]
atomic_valence[7] = [3, 4]
atomic_valence[8] = [2, 1, 3]
atomic_valence[9] = [1]
atomic_valence[14] = [4]
atomic_valence[15] = [5, 3]  # [5,4,3]
atomic_valence[16] = [6, 3, 2]  # [6,4,2]
atomic_valence[17] = [1]
atomic_valence[32] = [4]
atomic_valence[35] = [1]
atomic_valence[53] = [1]

atomic_valence_electrons = {}
atomic_valence_electrons[1] = 1
atomic_valence_electrons[5] = 3
atomic_valence_electrons[6] = 4
atomic_valence_electrons[7] = 5
atomic_valence_electrons[8] = 6
atomic_valence_electrons[9] = 7
atomic_valence_electrons[14] = 4
atomic_valence_electrons[15] = 5
atomic_valence_electrons[16] = 6
atomic_valence_electrons[17] = 7
atomic_valence_electrons[32] = 4
atomic_valence_electrons[35] = 7
atomic_valence_electrons[53] = 7


def str_atom(atom):
    """
    convert integer atom to string atom
    """
    global __ATOM_LIST__
    atom = __ATOM_LIST__[atom - 1]
    return atom


def int_atom(atom):
    """
    convert str atom to integer atom
    """
    global __ATOM_LIST__
    # print(atom)
    atom = atom.lower()
    return __ATOM_LIST__.index(atom) + 1


def get_UA(maxValence_list, valence_list):
    """ """
    UA = []
    DU = []
    for i, (maxValence, valence) in enumerate(zip(maxValence_list,
                                                  valence_list)):
        if not maxValence - valence > 0:
            continue
        UA.append(i)
        DU.append(maxValence - valence)
    return UA, DU


def get_BO(AC, UA, DU, valences, UA_pairs, use_graph=True):
    """ """
    BO = AC.copy()
    DU_save = []

    while DU_save != DU:
        for i, j in UA_pairs:
            BO[i, j] += 1
            BO[j, i] += 1

        BO_valence = list(BO.sum(axis=1))
        DU_save = copy.copy(DU)
        UA, DU = get_UA(valences, BO_valence)
        UA_pairs = get_UA_pairs(UA, AC, use_graph=use_graph)[0]

    return BO


def valences_not_too_large(BO, valences):
    """ """
    number_of_bonds_list = BO.sum(axis=1)
    for valence, number_of_bonds in zip(valences, number_of_bonds_list):
        if number_of_bonds > valence:
            return False

    return True


def charge_is_OK(
    BO,
    AC,
    charge,
    DU,
    atomic_valence_electrons,
    atoms,
    valences,
    allow_charged_fragments=True,
):
    # total charge
    Q = 0

    # charge fragment list
    q_list = []

    if allow_charged_fragments:

        BO_valences = list(BO.sum(axis=1))
        for i, atom in enumerate(atoms):
            q = get_atomic_charge(atom, atomic_valence_electrons[atom],
                                  BO_valences[i])
            Q += q
            if atom == 6:
                number_of_single_bonds_to_C = list(BO[i, :]).count(1)
                if number_of_single_bonds_to_C == 2 and BO_valences[i] == 2:
                    Q += 1
                    q = 2
                if number_of_single_bonds_to_C == 3 and Q + 1 < charge:
                    Q += 2
                    q = 1

            if q != 0:
                q_list.append(q)

    return charge == Q


def BO_is_OK(
    BO,
    AC,
    charge,
    DU,
    atomic_valence_electrons,
    atoms,
    valences,
    allow_charged_fragments=True,
):
    """
    Sanity of bond-orders
    args:
        BO -
        AC -
        charge -
        DU -
    optional
        allow_charges_fragments -
    returns:
        boolean - true of molecule is OK, false if not
    """

    if not valences_not_too_large(BO, valences):
        return False

    check_sum = (BO - AC).sum() == sum(DU)
    check_charge = charge_is_OK(
        BO,
        AC,
        charge,
        DU,
        atomic_valence_electrons,
        atoms,
        valences,
        allow_charged_fragments,
    )

    if check_charge and check_sum:
        return True

    return False


def get_atomic_charge(atom, atomic_valence_electrons, BO_valence):
    """ """

    if atom == 1:
        charge = 1 - BO_valence
    elif atom == 5:
        charge = 3 - BO_valence
    elif atom == 15 and BO_valence == 5:
        charge = 0
    elif atom == 16 and BO_valence == 6:
        charge = 0
    else:
        charge = atomic_valence_electrons - 8 + BO_valence

    return charge


def clean_charges(mol):
    """
    This hack should not be needed anymore, but is kept just in case
    """

    Chem.SanitizeMol(mol)
    # rxn_smarts = ['[N+:1]=[*:2]-[C-:3]>>[N+0:1]-[*:2]=[C-0:3]',
    #              '[N+:1]=[*:2]-[O-:3]>>[N+0:1]-[*:2]=[O-0:3]',
    #              '[N+:1]=[*:2]-[*:3]=[*:4]-[O-:5]>>[N+0:1]-[*:2]=[*:3]-[*:4]=[O-0:5]',
    #              '[#8:1]=[#6:2]([!-:6])[*:3]=[*:4][#6-:5]>>[*-:1][*:2]([*:6])=[*:3][*:4]=[*+0:5]',
    #              '[O:1]=[c:2][c-:3]>>[*-:1][*:2][*+0:3]',
    #              '[O:1]=[C:2][C-:3]>>[*-:1][*:2]=[*+0:3]']

    rxn_smarts = [
        "[#6,#7:1]1=[#6,#7:2][#6,#7:3]=[#6,#7:4][CX3-,NX3-:5][#6,#7:6]1=[#6,#7:7]>>"
        "[#6,#7:1]1=[#6,#7:2][#6,#7:3]=[#6,#7:4][-0,-0:5]=[#6,#7:6]1[#6-,#7-:7]",
        "[#6,#7:1]1=[#6,#7:2][#6,#7:3](=[#6,#7:4])[#6,#7:5]=[#6,#7:6][CX3-,NX3-:7]1>>"
        "[#6,#7:1]1=[#6,#7:2][#6,#7:3]([#6-,#7-:4])=[#6,#7:5][#6,#7:6]=[-0,-0:7]1",
    ]

    fragments = Chem.GetMolFrags(mol, asMols=True, sanitizeFrags=False)

    for i, fragment in enumerate(fragments):
        for smarts in rxn_smarts:
            patt = Chem.MolFromSmarts(smarts.split(">>")[0])
            while fragment.HasSubstructMatch(patt):
                rxn = AllChem.ReactionFromSmarts(smarts)
                ps = rxn.RunReactants((fragment,))
                fragment = ps[0][0]
                Chem.SanitizeMol(fragment)
        if i == 0:
            mol = fragment
        else:
            mol = Chem.CombineMols(mol, fragment)

    return mol


def BO2mol(
    mol,
    BO_matrix,
    atoms,
    atomic_valence_electrons,
    mol_charge,
    allow_charged_fragments=True,
):
    """
    based on code written by Paolo Toscani
    From bond order, atoms, valence structure and total charge, generate an
    rdkit molecule.
    args:
        mol - rdkit molecule
        BO_matrix - bond order matrix of molecule
        atoms - list of integer atomic symbols
        atomic_valence_electrons -
        mol_charge - total charge of molecule
    optional:
        allow_charged_fragments - bool - allow charged fragments
    returns
        mol - updated rdkit molecule with bond connectivity
    """

    l = len(BO_matrix)
    l2 = len(atoms)
    BO_valences = list(BO_matrix.sum(axis=1))

    if l != l2:
        raise RuntimeError(
            "sizes of adjMat ({0:d}) and Atoms {1:d} differ".format(l, l2))

    rwMol = Chem.RWMol(mol)

    bondTypeDict = {
        1: Chem.BondType.SINGLE,
        2: Chem.BondType.DOUBLE,
        3: Chem.BondType.TRIPLE,
    }

    for i in range(l):
        for j in range(i + 1, l):
            bo = int(round(BO_matrix[i, j]))
            if bo == 0:
                continue
            bt = bondTypeDict.get(bo, Chem.BondType.SINGLE)
            rwMol.AddBond(i, j, bt)

    mol = rwMol.GetMol()

    if allow_charged_fragments:
        mol = set_atomic_charges(mol, atoms, atomic_valence_electrons,
                                 BO_valences, BO_matrix, mol_charge)
    else:
        mol = set_atomic_radicals(mol, atoms, atomic_valence_electrons,
                                  BO_valences)

    return mol


def set_atomic_charges(mol, atoms, atomic_valence_electrons, BO_valences,
                       BO_matrix, mol_charge):
    """ """
    q = 0
    for i, atom in enumerate(atoms):
        a = mol.GetAtomWithIdx(i)
        charge = get_atomic_charge(atom, atomic_valence_electrons[atom],
                                   BO_valences[i])
        q += charge
        if atom == 6:
            number_of_single_bonds_to_C = list(BO_matrix[i, :]).count(1)
            if number_of_single_bonds_to_C == 2 and BO_valences[i] == 2:
                q += 1
                charge = 0
            if number_of_single_bonds_to_C == 3 and q + 1 < mol_charge:
                q += 2
                charge = 1

        if abs(charge) > 0:
            a.SetFormalCharge(int(charge))

    # mol = clean_charges(mol)

    return mol


def set_atomic_radicals(mol, atoms, atomic_valence_electrons, BO_valences):
    """
    The number of radical electrons = absolute atomic charge
    """
    for i, atom in enumerate(atoms):
        a = mol.GetAtomWithIdx(i)
        charge = get_atomic_charge(atom, atomic_valence_electrons[atom],
                                   BO_valences[i])

        if abs(charge) > 0:
            a.SetNumRadicalElectrons(abs(int(charge)))

    return mol


def get_bonds(UA, AC):
    """ """
    bonds = []

    for k, i in enumerate(UA):
        for j in UA[k + 1:]:
            if AC[i, j] == 1:
                bonds.append(tuple(sorted([i, j])))

    return bonds


def get_UA_pairs(UA, AC, use_graph=True):
    """ """

    bonds = get_bonds(UA, AC)

    if len(bonds) == 0:
        return [()]

    if use_graph:
        G = nx.Graph()
        G.add_edges_from(bonds)
        UA_pairs = [list(nx.max_weight_matching(G))]
        return UA_pairs

    max_atoms_in_combo = 0
    UA_pairs = [()]
    for combo in list(itertools.combinations(bonds, int(len(UA) / 2))):
        flat_list = [item for sublist in combo for item in sublist]
        atoms_in_combo = len(set(flat_list))
        if atoms_in_combo > max_atoms_in_combo:
            max_atoms_in_combo = atoms_in_combo
            UA_pairs = [combo]

        elif atoms_in_combo == max_atoms_in_combo:
            UA_pairs.append(combo)

    return UA_pairs


def AC2BO(AC, atoms, charge, allow_charged_fragments=True, use_graph=True):
    """
    implemenation of algorithm shown in Figure 2
    UA: unsaturated atoms
    DU: degree of unsaturation (u matrix in Figure)
    best_BO: Bcurr in Figure
    """

    global atomic_valence
    global atomic_valence_electrons

    # make a list of valences, e.g. for CO: [[4],[2,1]]
    valences_list_of_lists = []
    AC_valence = list(AC.sum(axis=1))

    for i, (atomicNum, valence) in enumerate(zip(atoms, AC_valence)):
        # valence can't be smaller than number of neighbourgs
        possible_valence = [
            x for x in atomic_valence[atomicNum] if x >= valence
        ]
        if not possible_valence:
            print_sys(
                "Valence of atom",
                i,
                "is",
                valence,
                "which bigger than allowed max",
                max(atomic_valence[atomicNum]),
                ". Stopping",
            )
            sys.exit()
        valences_list_of_lists.append(possible_valence)

    # convert [[4],[2,1]] to [[4,2],[4,1]]
    valences_list = itertools.product(*valences_list_of_lists)

    best_BO = AC.copy()

    for valences in valences_list:

        UA, DU_from_AC = get_UA(valences, AC_valence)

        check_len = len(UA) == 0
        if check_len:
            check_bo = BO_is_OK(
                AC,
                AC,
                charge,
                DU_from_AC,
                atomic_valence_electrons,
                atoms,
                valences,
                allow_charged_fragments=allow_charged_fragments,
            )
        else:
            check_bo = None

        if check_len and check_bo:
            return AC, atomic_valence_electrons

        UA_pairs_list = get_UA_pairs(UA, AC, use_graph=use_graph)
        for UA_pairs in UA_pairs_list:
            BO = get_BO(AC,
                        UA,
                        DU_from_AC,
                        valences,
                        UA_pairs,
                        use_graph=use_graph)
            status = BO_is_OK(
                BO,
                AC,
                charge,
                DU_from_AC,
                atomic_valence_electrons,
                atoms,
                valences,
                allow_charged_fragments=allow_charged_fragments,
            )
            charge_OK = charge_is_OK(
                BO,
                AC,
                charge,
                DU_from_AC,
                atomic_valence_electrons,
                atoms,
                valences,
                allow_charged_fragments=allow_charged_fragments,
            )

            if status:
                return BO, atomic_valence_electrons
            elif (BO.sum() >= best_BO.sum() and
                  valences_not_too_large(BO, valences) and charge_OK):
                best_BO = BO.copy()

    return best_BO, atomic_valence_electrons


def AC2mol(mol,
           AC,
           atoms,
           charge,
           allow_charged_fragments=True,
           use_graph=True):
    """ """

    # convert AC matrix to bond order (BO) matrix
    BO, atomic_valence_electrons = AC2BO(
        AC,
        atoms,
        charge,
        allow_charged_fragments=allow_charged_fragments,
        use_graph=use_graph,
    )

    # add BO connectivity and charge info to mol object
    mol = BO2mol(
        mol,
        BO,
        atoms,
        atomic_valence_electrons,
        charge,
        allow_charged_fragments=allow_charged_fragments,
    )
    # print('mol', mol)
    # If charge is not correct don't return mol
    # if Chem.GetFormalCharge(mol) != charge:
    #     return []

    # BO2mol returns an arbitrary resonance form. Let's make the rest
    mols = rdchem.ResonanceMolSupplier(mol, Chem.UNCONSTRAINED_CATIONS,
                                       Chem.UNCONSTRAINED_ANIONS)
    mols = [mol for mol in mols]

    return mols, BO


def get_proto_mol(atoms):
    """ """
    mol = Chem.MolFromSmarts("[#" + str(atoms[0]) + "]")
    rwMol = Chem.RWMol(mol)
    for i in range(1, len(atoms)):
        a = Chem.Atom(atoms[i])
        rwMol.AddAtom(a)

    mol = rwMol.GetMol()

    return mol


def read_xyz_file(filename, look_for_charge=True):
    """
    from https://github.com/jensengroup/xyz2mol/blob/master/xyz2mol.py
    """

    atomic_symbols = []
    xyz_coordinates = []
    charge = 0
    title = ""

    with open(filename, "r") as file:
        for line_number, line in enumerate(file):
            if line_number == 0:
                num_atoms = int(line)
            elif line_number == 1:
                title = line
                if "charge=" in line:
                    charge = int(line.split("=")[1])
            else:
                atomic_symbol, x, y, z = line.split()
                atomic_symbols.append(atomic_symbol)
                xyz_coordinates.append([float(x), float(y), float(z)])

    atoms = [int_atom(atom) for atom in atomic_symbols]

    return atoms, charge, xyz_coordinates


def xyz2AC(atoms, xyz, charge, use_huckel=False):
    """
    atoms and coordinates to atom connectivity (AC)
    args:
        atoms - int atom types
        xyz - coordinates
        charge - molecule charge
    optional:
        use_huckel - Use Huckel method for atom connecitivty
    returns
        ac - atom connectivity matrix
        mol - rdkit molecule
    """

    if use_huckel:
        return xyz2AC_huckel(atoms, xyz, charge)
    else:
        return xyz2AC_vdW(atoms, xyz)


def xyz2AC_vdW(atoms, xyz):

    # Get mol template
    mol = get_proto_mol(atoms)

    # Set coordinates
    conf = Chem.Conformer(mol.GetNumAtoms())
    for i in range(mol.GetNumAtoms()):
        conf.SetAtomPosition(i, (xyz[i][0], xyz[i][1], xyz[i][2]))
    mol.AddConformer(conf)

    AC = get_AC(mol)

    return AC, mol


def get_AC(mol, covalent_factor=1.3):
    """
    Generate adjacent matrix from atoms and coordinates.
    AC is a (num_atoms, num_atoms) matrix with 1 being covalent bond and 0 is not
    covalent_factor - 1.3 is an arbitrary factor
    args:
        mol - rdkit molobj with 3D conformer
    optional
        covalent_factor - increase covalent bond length threshold with facto
    returns:
        AC - adjacent matrix
    """

    # Calculate distance matrix
    dMat = Chem.Get3DDistanceMatrix(mol)

    pt = Chem.GetPeriodicTable()
    num_atoms = mol.GetNumAtoms()
    AC = np.zeros((num_atoms, num_atoms), dtype=int)

    for i in range(num_atoms):
        a_i = mol.GetAtomWithIdx(i)
        Rcov_i = pt.GetRcovalent(a_i.GetAtomicNum()) * covalent_factor
        for j in range(i + 1, num_atoms):
            a_j = mol.GetAtomWithIdx(j)
            Rcov_j = pt.GetRcovalent(a_j.GetAtomicNum()) * covalent_factor
            if dMat[i, j] <= Rcov_i + Rcov_j:
                AC[i, j] = 1
                AC[j, i] = 1

    return AC


def xyz2AC_huckel(atomicNumList, xyz, charge):
    """
    args
        atomicNumList - atom type list
        xyz - coordinates
        charge - molecule charge
    returns
        ac - atom connectivity
        mol - rdkit molecule
    """
    mol = get_proto_mol(atomicNumList)

    conf = Chem.Conformer(mol.GetNumAtoms())
    for i in range(mol.GetNumAtoms()):
        conf.SetAtomPosition(i, (xyz[i][0], xyz[i][1], xyz[i][2]))
    mol.AddConformer(conf)

    num_atoms = len(atomicNumList)
    AC = np.zeros((num_atoms, num_atoms)).astype(int)

    mol_huckel = Chem.Mol(mol)
    mol_huckel.GetAtomWithIdx(0).SetFormalCharge(
        charge)  # mol charge arbitrarily added to 1st atom

    passed, result = rdEHTTools.RunMol(mol_huckel)
    opop = result.GetReducedOverlapPopulationMatrix()
    tri = np.zeros((num_atoms, num_atoms))
    tri[np.tril(np.ones(
        (num_atoms, num_atoms),
        dtype=bool))] = opop  # lower triangular to square matrix
    for i in range(num_atoms):
        for j in range(i + 1, num_atoms):
            pair_pop = abs(tri[j, i])
            if pair_pop >= 0.15:  # arbitry cutoff for bond. May need adjustment
                AC[i, j] = 1
                AC[j, i] = 1

    return AC, mol


def chiral_stereo_check(mol):
    """
    Find and embed chiral information into the model based on the coordinates
    args:
        mol - rdkit molecule, with embeded conformer
    """
    Chem.SanitizeMol(mol)
    Chem.DetectBondStereochemistry(mol, -1)
    Chem.AssignStereochemistry(mol, flagPossibleStereoCenters=True, force=True)
    Chem.AssignAtomChiralTagsFromStructure(mol, -1)

    return


def xyz2mol(
    atoms,
    coordinates,
    charge=0,
    allow_charged_fragments=True,
    use_graph=True,
    use_huckel=False,
    embed_chiral=True,
):
    """
    Generate a rdkit molobj from atoms, coordinates and a total_charge.
    args:
        atoms - list of atom types (int)
        coordinates - 3xN Cartesian coordinates
        charge - total charge of the system (default: 0)
    optional:
        allow_charged_fragments - alternatively radicals are made
        use_graph - use graph (networkx)
        use_huckel - Use Huckel method for atom connectivity prediction
        embed_chiral - embed chiral information to the molecule
    returns:
        mols - list of rdkit molobjects
    """

    # Get atom connectivity (AC) matrix, list of atomic numbers, molecular charge,
    # and mol object with no connectivity information
    AC, mol = xyz2AC(atoms, coordinates, charge, use_huckel=use_huckel)
    # print('AC, mol', AC, mol)
    # Convert AC to bond order matrix and add connectivity and charge info to
    # mol object
    new_mols, BO = AC2mol(
        mol,
        AC,
        atoms,
        charge,
        allow_charged_fragments=allow_charged_fragments,
        use_graph=use_graph,
    )
    # Check for stereocenters and chiral centers
    # print('new_mols', new_mols)
    if embed_chiral:
        for new_mol in new_mols:
            chiral_stereo_check(new_mol)
    return new_mols, BO


def xyzfile2mol(xyzfile):
    atoms, charge, xyz_coordinates = read_xyz_file(xyzfile)
    # print(atoms, xyz_coordinates)
    charged_fragments = True
    # quick is faster for large systems but requires networkx
    # if you don't want to install networkx set quick=False and
    # uncomment 'import networkx as nx' at the top of the file
    quick = True

    # chiral comment
    embed_chiral = True  ### random

    # huckel uses extended Huckel bond orders to locate bonds (requires RDKit 2019.9.1 or later)
    # otherwise van der Waals radii are used
    use_huckel = True

    # if explicit charge from args, set it
    charge = charge

    # Get the molobjs
    mols, BO = xyz2mol(
        atoms,
        xyz_coordinates,
        charge=charge,
        use_graph=quick,
        allow_charged_fragments=charged_fragments,
        embed_chiral=embed_chiral,
        use_huckel=use_huckel,
    )
    return mols[0], BO



"""Utilities functions for splitting dataset 
"""
import os, sys
import numpy as np
import pandas as pd
from tqdm import tqdm
from .misc import print_sys


def create_fold(df, fold_seed, frac):
    """create random split

    Args:
        df (pd.DataFrame): dataset dataframe
        fold_seed (int): the random seed
        frac (list): a list of train/valid/test fractions

    Returns:
        dict: a dictionary of splitted dataframes, where keys are train/valid/test and values correspond to each dataframe
    """
    train_frac, val_frac, test_frac = frac
    test = df.sample(frac=test_frac, replace=False, random_state=fold_seed)
    train_val = df[~df.index.isin(test.index)]
    val = train_val.sample(frac=val_frac / (1 - test_frac),
                           replace=False,
                           random_state=1)
    train = train_val[~train_val.index.isin(val.index)]

    return {
        "train": train.reset_index(drop=True),
        "valid": val.reset_index(drop=True),
        "test": test.reset_index(drop=True),
    }


def create_fold_setting_cold(df, fold_seed, frac, entities):
    """create cold-split where given one or multiple columns, it first splits based on
    entities in the columns and then maps all associated data points to the partition

    Args:
            df (pd.DataFrame): dataset dataframe
            fold_seed (int): the random seed
            frac (list): a list of train/valid/test fractions
            entities (Union[str, List[str]]): either a single "cold" entity or a list of
                    "cold" entities on which the split is done

    Returns:
            dict: a dictionary of splitted dataframes, where keys are train/valid/test and values correspond to each dataframe
    """
    if isinstance(entities, str):
        entities = [entities]

    train_frac, val_frac, test_frac = frac

    # For each entity, sample the instances belonging to the test datasets
    test_entity_instances = [
        df[e].drop_duplicates().sample(frac=test_frac,
                                       replace=False,
                                       random_state=fold_seed).values
        for e in entities
    ]

    # Select samples where all entities are in the test set
    test = df.copy()
    for entity, instances in zip(entities, test_entity_instances):
        test = test[test[entity].isin(instances)]

    if len(test) == 0:
        raise ValueError(
            "No test samples found. Try another seed, increasing the test frac or a "
            "less stringent splitting strategy.")

    # Proceed with validation data
    train_val = df.copy()
    for i, e in enumerate(entities):
        train_val = train_val[~train_val[e].isin(test_entity_instances[i])]

    val_entity_instances = [
        train_val[e].drop_duplicates().sample(frac=val_frac / (1 - test_frac),
                                              replace=False,
                                              random_state=fold_seed).values
        for e in entities
    ]
    val = train_val.copy()
    for entity, instances in zip(entities, val_entity_instances):
        val = val[val[entity].isin(instances)]

    if len(val) == 0:
        raise ValueError(
            "No validation samples found. Try another seed, increasing the test frac "
            "or a less stringent splitting strategy.")

    train = train_val.copy()
    for i, e in enumerate(entities):
        train = train[~train[e].isin(val_entity_instances[i])]

    return {
        "train": train.reset_index(drop=True),
        "valid": val.reset_index(drop=True),
        "test": test.reset_index(drop=True),
    }


def create_scaffold_split(df, seed, frac, entity):
    """create scaffold split. it first generates molecular scaffold for each molecule and then split based on scaffolds
    reference: https://github.com/chemprop/chemprop/blob/master/chemprop/data/scaffold.py

    Args:
        df (pd.DataFrame): dataset dataframe
        fold_seed (int): the random seed
        frac (list): a list of train/valid/test fractions
        entity (str): the column name for where molecule stores

    Returns:
        dict: a dictionary of splitted dataframes, where keys are train/valid/test and values correspond to each dataframe
    """

    try:
        from rdkit import Chem
        from rdkit.Chem.Scaffolds import MurckoScaffold
        from rdkit import RDLogger

        RDLogger.DisableLog("rdApp.*")
    except:
        raise ImportError(
            "Please install rdkit by 'conda install -c conda-forge rdkit'! ")
    from tqdm import tqdm
    from random import Random

    from collections import defaultdict

    random = Random(seed)

    s = df[entity].values
    scaffolds = defaultdict(set)
    idx2mol = dict(zip(list(range(len(s))), s))

    error_smiles = 0
    for i, smiles in tqdm(enumerate(s), total=len(s)):
        try:
            scaffold = MurckoScaffold.MurckoScaffoldSmiles(
                mol=Chem.MolFromSmiles(smiles), includeChirality=False)
            scaffolds[scaffold].add(i)
        except:
            print_sys(smiles + " returns RDKit error and is thus omitted...")
            error_smiles += 1

    train, val, test = [], [], []
    train_size = int((len(df) - error_smiles) * frac[0])
    val_size = int((len(df) - error_smiles) * frac[1])
    test_size = (len(df) - error_smiles) - train_size - val_size
    train_scaffold_count, val_scaffold_count, test_scaffold_count = 0, 0, 0

    # index_sets = sorted(list(scaffolds.values()), key=lambda i: len(i), reverse=True)
    index_sets = list(scaffolds.values())
    big_index_sets = []
    small_index_sets = []
    for index_set in index_sets:
        if len(index_set) > val_size / 2 or len(index_set) > test_size / 2:
            big_index_sets.append(index_set)
        else:
            small_index_sets.append(index_set)
    random.seed(seed)
    random.shuffle(big_index_sets)
    random.shuffle(small_index_sets)
    index_sets = big_index_sets + small_index_sets

    if frac[2] == 0:
        for index_set in index_sets:
            if len(train) + len(index_set) <= train_size:
                train += index_set
                train_scaffold_count += 1
            else:
                val += index_set
                val_scaffold_count += 1
    else:
        for index_set in index_sets:
            if len(train) + len(index_set) <= train_size:
                train += index_set
                train_scaffold_count += 1
            elif len(val) + len(index_set) <= val_size:
                val += index_set
                val_scaffold_count += 1
            else:
                test += index_set
                test_scaffold_count += 1

    return {
        "train": df.iloc[train].reset_index(drop=True),
        "valid": df.iloc[val].reset_index(drop=True),
        "test": df.iloc[test].reset_index(drop=True),
    }


def create_combination_generation_split(dict1, dict2, seed, frac):
    """create random split

    Args:
        dict: data dict
        fold_seed (int): the random seed
        frac (list): a list of train/valid/test fractions

    Returns:
        dict: a dictionary of splitted dataframes, where keys are train/valid/test and values correspond to each dataframe
    """
    train_frac, val_frac, test_frac = frac
    length = len(dict1["coord"])
    indices = np.random.permutation(length)
    train_idx, val_idx, test_idx = (
        indices[:int(length * train_frac)],
        indices[int(length * train_frac):int(length * (train_frac + val_frac))],
        indices[int(length * (train_frac + val_frac)):],
    )

    return {
        "train": {
            "protein_coord": [dict1["coord"][i] for i in train_idx],
            "protein_atom_type": [dict1["atom_type"][i] for i in train_idx],
            "ligand_coord": [dict2["coord"][i] for i in train_idx],
            "ligand_atom_type": [dict2["atom_type"][i] for i in train_idx],
        },
        "valid": {
            "protein_coord": [dict1["coord"][i] for i in val_idx],
            "protein_atom_type": [dict1["atom_type"][i] for i in val_idx],
            "ligand_coord": [dict2["coord"][i] for i in val_idx],
            "ligand_atom_type": [dict2["atom_type"][i] for i in val_idx],
        },
        "test": {
            "protein_coord": [dict1["coord"][i] for i in test_idx],
            "protein_atom_type": [dict1["atom_type"][i] for i in test_idx],
            "ligand_coord": [dict2["coord"][i] for i in test_idx],
            "ligand_atom_type": [dict2["atom_type"][i] for i in test_idx],
        },
    }


def create_combination_split(df, seed, frac):
    """
    Function for splitting drug combination dataset such that no combinations are shared across the split

    Args:
        df (pd.Dataframe): dataset to split
        seed (int): random seed
        frac (list): split fraction as a list

    Returns:
        dict: a dictionary of splitted dataframes, where keys are train/valid/test and values correspond to each dataframe
    """

    test_size = int(len(df) * frac[2])
    train_size = int(len(df) * frac[0])
    val_size = len(df) - train_size - test_size
    np.random.seed(seed)

    # Create a new column for combination names
    df["concat"] = df["Drug1_ID"] + "," + df["Drug2_ID"]

    # Identify shared drug combinations across all target classes
    combinations = []
    for c in df["Cell_Line_ID"].unique():
        df_cell = df[df["Cell_Line_ID"] == c]
        combinations.append(set(df_cell["concat"].values))

    intxn = combinations[0]
    for c in combinations:
        intxn = intxn.intersection(c)

    # Split combinations into train, val and test
    test_choices = np.random.choice(list(intxn),
                                    int(test_size /
                                        len(df["Cell_Line_ID"].unique())),
                                    replace=False)
    trainval_intxn = intxn.difference(test_choices)
    val_choices = np.random.choice(
        list(trainval_intxn),
        int(val_size / len(df["Cell_Line_ID"].unique())),
        replace=False,
    )

    ## Create train and test set
    test_set = df[df["concat"].isin(test_choices)].drop(columns=["concat"])
    val_set = df[df["concat"].isin(val_choices)]
    train_set = df[~df["concat"].isin(test_choices)].reset_index(drop=True)
    train_set = train_set[~train_set["concat"].isin(val_choices)]

    return {
        "train": train_set.reset_index(drop=True),
        "valid": val_set.reset_index(drop=True),
        "test": test_set.reset_index(drop=True),
    }


# create time split


def create_fold_time(df, frac, date_column):
    """create splits based on time

    Args:
        df (pd.DataFrame): the dataset dataframe
        frac (list): list of train/valid/test fractions
        date_column (str): the name of the column that contains the time info

    Returns:
        dict: a dictionary of splitted dataframes, where keys are train/valid/test and values correspond to each dataframe
    """
    df = df.sort_values(by=date_column).reset_index(drop=True)
    train_frac, val_frac, test_frac = frac[0], frac[1], frac[2]

    split_date = df[:int(len(df) *
                         (train_frac + val_frac))].iloc[-1][date_column]
    test = df[df[date_column] >= split_date].reset_index(drop=True)
    train_val = df[df[date_column] < split_date]

    split_date_valid = train_val[:int(
        len(train_val) * train_frac /
        (train_frac + val_frac))].iloc[-1][date_column]
    train = train_val[train_val[date_column] <= split_date_valid].reset_index(
        drop=True)
    valid = train_val[train_val[date_column] > split_date_valid].reset_index(
        drop=True)

    return {
        "train": train,
        "valid": valid,
        "test": test,
        "split_time": {
            "train_time_frame": (df.iloc[0][date_column], split_date_valid),
            "valid_time_frame": (split_date_valid, split_date),
            "test_time_frame": (split_date, df.iloc[-1][date_column]),
        },
    }


def create_group_split(train_val, seed, holdout_frac, group_column):
    """split within each stratification defined by the group column for training/validation split

    Args:
        train_val (pd.DataFrame): the train+valid dataframe to split on
        seed (int): the random seed
        holdout_frac (float): the fraction of validation
        group_column (str): the name of the group column

    Returns:
        dict: a dictionary of splitted dataframes, where keys are train/valid/test and values correspond to each dataframe
    """
    train_df = pd.DataFrame()
    val_df = pd.DataFrame()

    for i in train_val[group_column].unique():
        train_val_temp = train_val[train_val[group_column] == i]
        np.random.seed(seed)
        msk = np.random.rand(len(train_val_temp)) < (1 - holdout_frac)
        train_df = pd.concat([train_df, train_val_temp[msk]], axis=0)
        val_df = pd.concat([val_df, train_val_temp[~msk]], axis=0)

    return {
        "train": train_df.reset_index(drop=True),
        "valid": val_df.reset_index(drop=True),
    }



import cellxgene_census
import gget
import pandas as pd
from scipy.sparse import csr_matrix
import tiledbsoma


class CensusResource:

    _CENSUS_DATA = "census_data"
    _CENSUS_META = "census_info"
    _FEATURE_PRESENCE = "feature_dataset_presence_matrix"
    _LATEST_CENSUS = "stable"
    _HUMAN = "homo_sapiens"

    class decorators:

        @classmethod
        def check_dataset_is_census_data(cls, func):
            """Sets self.dataset to census_data"""

            def check(*args, **kwargs):
                self = args[0]
                self.dataset = self._CENSUS_DATA
                return func(*args, **kwargs)

            return check

        @classmethod
        def check_dataset_is_census_info(cls, func):
            """Sets self.dataset to census_data"""

            def check(*args, **kwargs):
                self = args[0]
                self.dataset = self._CENSUS_META
                return func(*args, **kwargs)

            return check

        @classmethod
        def slice_checks_X_and_FM(cls, func):
            """Decorator for:
            1. functions that need X and feature presence matrix apply slicing if not filtering
            2. functions with a todense() option abide by required formatting
            3. functions requiring a measurement name provide a measurement name
            4. fmt is a valid format
            asserts these requirements hold in input arguments."""

            def check(*args, **kwargs):
                if "upper" in kwargs:
                    upper, lower = kwargs.get('upper',
                                              None), kwargs.get("lower", None)
                    if upper is None or lower is None:
                        raise Exception(
                            "No upper and/or lower bound for slicing was provided. Dataset is too large to fit in memory. \
                            Memory-Efficient queries are not supported yet.")
                fmt = kwargs.get("fmt")
                fmt = fmt if fmt is not None else "pandas"
                if "todense" in kwargs:
                    todense = kwargs.get("todense")
                    kwargs[
                        "todense"] = todense if todense is not None else False
                    if todense and fmt != "scipy":
                        raise ValueError(
                            "dense representation only available in scipy format"
                        )
                measurement_name = kwargs.get("measurement_name")
                if measurement_name is None:
                    raise Exception("measurement_name was not provided.")
                kwargs["fmt"] = fmt if fmt is not None else "pandas"
                return func(*args, **kwargs)

            return check

    def __init__(self, census_version=None, organism=None):
        """Initialize the Census Resource.
        
        Args:
            census_version (str): The date of the census data release to use 
            organism (str): string for census data organism to query data for. defaults to human.
        """
        self.census_version = census_version if census_version is not None else self._LATEST_CENSUS
        self.organism = organism if organism is not None else self._HUMAN
        self.dataset = None  # variable to set target census collection to either info or data

    def fmt_cellxgene_data(self,
                           tiledb_ptr,
                           fmt=None,
                           input_is_table=False,
                           is_csr=False,
                           is_soma=False):
        """Transform TileDB DataFrame or SparseNDArray to one of the supported API formats.

        Args:
            tiledb_ptr (TileDB DataFrame or SparseNDArray): pointer to the TileDB DataFrame
            fmt (str, optional): deisgnates a format to transfowm TileDB data to. Defaults to None.

        Raises:
            Exception: if no format is provided
            Exception: if format is not a valid option

        Returns:
            The dataset in selected format if it's a valid format
        """
        if fmt is None:
            raise Exception(
                "format not provided to fmt_cellxgene_data(), please provide fmt variable"
            )
        elif is_soma:
            import tiledb
            uri = tiledb_ptr.uri
            config = tiledb.Config({
                'vfs.s3.region': 'us-west-2',
            })
            ctx = tiledb.Ctx(config=config)
            with tiledb.open(uri, ctx=ctx, mode="r") as df:
                return pd.DataFrame(df[:])
        elif fmt == "pandas" and is_csr:
            return pd.DataFrame(tiledb_ptr.toarray())
        elif fmt == "pandas" and not input_is_table:
            return tiledb_ptr.concat().to_pandas()
        elif fmt == "pandas":
            return tiledb_ptr.to_pandas()
        elif fmt == "pyarrow" and not input_is_table:
            return tiledb_ptr.concat()
        elif fmt == "pyarrow":
            return tiledb_ptr
        elif fmt == "scipy" and not input_is_table:
            return tiledb_ptr.concat().to_scipy()
        elif fmt == "scipy":
            return csr_matrix(tiledb_ptr.to_pandas().values)
        else:
            raise Exception(
                "fmt not in [pandas, pyarrow, scipy] for fmt_cellxgene_data()")

    @decorators.check_dataset_is_census_data
    def get_cell_metadata(self, value_filter=None, column_names=None, fmt=None):
        """Get the cell metadata (obs) data from the Census API"""
        if value_filter is None:
            raise Exception(
                "No value filter was provided, dataset is too large to fit in memory. \
                            Memory-Efficient queries are not supported yet.")
        fmt = fmt if fmt is not None else "pandas"
        with cellxgene_census.open_soma(
                census_version=self.census_version) as census:
            obs = census[self.dataset][self.organism].obs
            obsread = None
            if column_names:
                obsread = obs.read(value_filter=value_filter,
                                   column_names=column_names)
            else:
                obsread = obs.read(value_filter=value_filter)
            return self.fmt_cellxgene_data(obsread, fmt)

    @decorators.check_dataset_is_census_data
    def get_gene_metadata(self,
                          value_filter=None,
                          column_names=None,
                          measurement_name=None,
                          fmt=None):
        """Get the gene metadata (var) data from the Census API"""
        if value_filter is None:
            raise Exception(
                "No value filter was provided, dataset is too large to fit in memory. \
                Memory-Efficient queries are not supported yet.")
        elif measurement_name is None:
            raise ValueError("measurment_name must be provided , i.e. 'RNA'")
        fmt = fmt if fmt is not None else "pandas"
        with cellxgene_census.open_soma(
                census_version=self.census_version) as census:
            var = census[self.dataset][self.organism].ms[measurement_name].var
            varread = None
            if column_names:
                varread = var.read(value_filter=value_filter,
                                   column_names=column_names)
            else:
                varread = var.read(value_filter=value_filter)
            return self.fmt_cellxgene_data(varread, fmt)

    @decorators.slice_checks_X_and_FM
    @decorators.check_dataset_is_census_data
    def get_measurement_matrix(self,
                               upper=None,
                               lower=None,
                               value_adjustment=None,
                               measurement_name=None,
                               fmt=None,
                               todense=None):
        """Count matrix for an input measurement by slice

        Args:
            upper (int, optional): upper bound on the slice to obtain. Defaults to None.
            lower (int, optional): lower bound on the slice to obtain. Defaults to None.
            value_adjustment (str, optional): designates the type of count desired for this measurement. Defaults to None.
            measurement_name (str, optional): name of measurement, i.e. 'raw'. Defaults to None.

        Returns:
            A slice from the count matrix in the specified format. If `todense` is True, then a dense scipy array will be returned.
        """
        value_adjustment = value_adjustment if value_adjustment is not None else "raw"
        with cellxgene_census.open_soma(
                census_version=self.census_version) as census:
            n_obs = len(census[self.dataset][self.organism].obs)
            n_var = len(
                census[self.dataset][self.organism].ms[measurement_name].var)
            X = census[self.dataset][
                self.organism].ms[measurement_name].X[value_adjustment]
            slc = X.read([slice(lower, upper)]).coos((n_obs, n_var))
            out = self.fmt_cellxgene_data(slc, fmt)
            return out if not todense else out.todense()

    @decorators.slice_checks_X_and_FM
    @decorators.check_dataset_is_census_data
    def get_feature_dataset_presence_matrix(self,
                                            upper=None,
                                            lower=None,
                                            measurement_name=None,
                                            fmt=None,
                                            todense=None):
        """Gets a slice from the feature_dataset_presence_matrix for a given measurement_name

        Args:
            upper (int, optional): upper bound on the slice. Defaults to None.
            lower (int, optional): lower bound on the slice. Defaults to None.
            measurement_name (str, optional): measurment_name for the query i.e. 'rna'. Defaults to None.
            fmt (str, optional): output format desired for the output dataset. Defaults to None.
            todense (bool, optional): if True, returns scipy dense representation. Defaults to None.

        Returns:
            dataset in desired format
        """
        with cellxgene_census.open_soma(
                census_version=self.census_version) as census:
            n_obs = len(census[self.dataset][self.organism].obs)
            n_var = len(
                census[self.dataset][self.organism].ms[measurement_name].var)
            fMatrix = census[self.dataset][self.organism].ms[measurement_name][
                "feature_dataset_presence_matrix"]
            slc = fMatrix.read((slice(lower, upper),)).coos((n_obs, n_var))
            out = self.fmt_cellxgene_data(slc, fmt)
            return out if not todense else out.todense()

    @decorators.check_dataset_is_census_data
    def get_feature_dataset_presence_matrix_entry(self,
                                                  dataset_name,
                                                  measurement_name=None,
                                                  fmt=None,
                                                  todense=None):
        with cellxgene_census.open_soma(
                census_version=self.census_version) as census:
            fMatrix = cellxgene_census.get_presence_matrix(
                census,
                organism=self.organism,
                measurement_name=measurement_name)
            meta_df = self.get_dataset_metadata()
            dataset_id = meta_df[meta_df["dataset_title"] == dataset_name].index
            entry = fMatrix[dataset_id]
            out = self.fmt_cellxgene_data(entry, "pandas", is_csr=True)
            return out if not todense else out.todense()

    @decorators.check_dataset_is_census_info
    def get_metadata(self):
        """Get the metadata for the Cell Census."""
        with cellxgene_census.open_soma(
                census_version=self.census_version) as census:
            return census[self.dataset]["summary"]

    @decorators.check_dataset_is_census_info
    def get_dataset_metadata(self):
        """Get the metadata for the Cell Census's datasets."""
        with cellxgene_census.open_soma(
                census_version=self.census_version) as census:
            return self.fmt_cellxgene_data(census[self.dataset]["datasets"],
                                           fmt="pandas",
                                           is_soma=True)

    @decorators.check_dataset_is_census_info
    def get_cell_count_metadata(self):
        """Get the cell counts across cell metadata for the Cell Census."""
        with cellxgene_census.open_soma(
                census_version=self.census_version) as census:
            return census[self.dataset]["summary_cell_counts"]

    @decorators.slice_checks_X_and_FM
    @decorators.check_dataset_is_census_data
    def query_measurement_matrix(self,
                                 value_filter=None,
                                 value_adjustment=None,
                                 measurement_name=None,
                                 fmt=None,
                                 todense=None,
                                 gene_filter=None):
        """Query the Census Measurement Matrix. Function returns a Python generator.

        Args:
            value_filter (str, optional): a valuer filter (obs) to apply to the query. Defaults to None.
            value_adjustment (str, optional): the type of count to obtain from count matricx for this measurement. Defaults to None.
            measurement_name (str, optional): measurement name to query, i.e. "RNA". Defaults to None.
            fmt (str, optional): output format for the output dataset. Defaults to None.
            todense (bool, optional): if True, will output a dense scipy array as the representation. Defaults to None.

        Yields:
            a slice of the output query in the specified format
        """
        import numpy
        value_adjustment = value_adjustment if value_adjustment is not None else "raw"
        with cellxgene_census.open_soma(
                census_version=self.census_version) as census:
            organism = census[self.dataset][self.organism]
            query = organism.axis_query(
                measurement_name=measurement_name,
                obs_query=tiledbsoma.AxisQuery(value_filter=value_filter)
                if value_filter else tiledbsoma.AxisQuery(),
                var_query=tiledbsoma.AxisQuery(
                    coords=(numpy.array(gene_filter),))
                if gene_filter else tiledbsoma.AxisQuery(),
            )
            it = query.X(value_adjustment).tables()
            for slc in it:
                out = self.fmt_cellxgene_data(slc, fmt, input_is_table=True)
                out = out if not todense else out.todense()
                yield out

    @classmethod
    def gget_czi_cellxgene(cls, **kwargs):
        """Wrapper for cellxgene gget()
            https://chanzuckerberg.github.io/cellxgene-census/notebooks/api_demo/census_gget_demo.html
            Support for AnnData or DataFrame. Params included below
            
            General args:
            - species        Choice of 'homo_sapiens' or 'mus_musculus'. Default: 'homo_sapiens'.
            - gene           Str or list of gene name(s) or Ensembl ID(s), e.g. ['ACE2', 'SLC5A1'] or ['ENSG00000130234', 'ENSG00000100170']. Default: None.
                            NOTE: Set ensembl=True when providing Ensembl ID(s) instead of gene name(s).
                            See https://cellxgene.cziscience.com/gene-expression for examples of available genes.
            - ensembl        True/False (default: False). Set to True when genes are provided as Ensembl IDs.
            - column_names   List of metadata columns to return (stored in AnnData.obs when meta_only=False).
                            Default: ["dataset_id", "assay", "suspension_type", "sex", "tissue_general", "tissue", "cell_type"]
                            For more options see: https://api.cellxgene.cziscience.com/curation/ui/#/ -> Schemas -> dataset
            - meta_only      True/False (default: False). If True, returns only metadata dataframe (corresponds to AnnData.obs).
            - census_version Str defining version of Census, e.g. "2023-05-15" or "latest" or "stable". Default: "stable".
            - verbose        True/False whether to print progress information. Default True.
            - out            If provided, saves the generated AnnData h5ad (or csv when meta_only=True) file with the specified path. Default: None.

        Cell metadata attributes:
            - tissue                          Str or list of tissue(s), e.g. ['lung', 'blood']. Default: None.
                                            See https://cellxgene.cziscience.com/gene-expression for examples of available tissues.
            - cell_type                       Str or list of celltype(s), e.g. ['mucus secreting cell', 'neuroendocrine cell']. Default: None.
                                            See https://cellxgene.cziscience.com/gene-expression and select a tissue to see examples of available celltypes.
            - development_stage               Str or list of development stage(s). Default: None.
            - disease                         Str or list of disease(s). Default: None.
            - sex                             Str or list of sex(es), e.g. 'female'. Default: None.
            - is_primary_data                 True/False (default: True). If True, returns only the canonical instance of the cellular observation.
                                            This is commonly set to False for meta-analyses reusing data or for secondary views of data.
            - dataset_id                      Str or list of CELLxGENE dataset ID(s). Default: None.
            - tissue_general_ontology_term_id Str or list of high-level tissue UBERON ID(s). Default: None.
                                            Also see: https://github.com/chanzuckerberg/single-cell-data-portal/blob/9b94ccb0a2e0a8f6182b213aa4852c491f6f6aff/backend/wmg/data/tissue_mapper.py
            - tissue_general                  Str or list of high-level tissue label(s). Default: None.
                                            Also see: https://github.com/chanzuckerberg/single-cell-data-portal/blob/9b94ccb0a2e0a8f6182b213aa4852c491f6f6aff/backend/wmg/data/tissue_mapper.py
            - tissue_ontology_term_id         Str or list of tissue ontology term ID(s) as defined in the CELLxGENE dataset schema. Default: None.
            - assay_ontology_term_id          Str or list of assay ontology term ID(s) as defined in the CELLxGENE dataset schema. Default: None.
            - assay                           Str or list of assay(s) as defined in the CELLxGENE dataset schema. Default: None.
            - cell_type_ontology_term_id      Str or list of celltype ontology term ID(s) as defined in the CELLxGENE dataset schema. Default: None.
            - development_stage_ontology_term_id        Str or list of development stage ontology term ID(s) as defined in the CELLxGENE dataset schema. Default: None.
            - disease_ontology_term_id        Str or list of disease ontology term ID(s) as defined in the CELLxGENE dataset schema. Default: None.
            - donor_id                        Str or list of donor ID(s) as defined in the CELLxGENE dataset schema. Default: None.
            - self_reported_ethnicity_ontology_term_id  Str or list of self reported ethnicity ontology ID(s) as defined in the CELLxGENE dataset schema. Default: None.
            - self_reported_ethnicity         Str or list of self reported ethnicity as defined in the CELLxGENE dataset schema. Default: None.
            - sex_ontology_term_id            Str or list of sex ontology ID(s) as defined in the CELLxGENE dataset schema. Default: None.
            - suspension_type                 Str or list of suspension type(s) as defined in the CELLxGENE dataset schema. Default: None.

        Returns AnnData object (when meta_only=False) or dataframe (when meta_only=True).

        """
        gget.setup("cellxgene")
        return gget.cellxgene(**kwargs)

    def get_anndata(self, add_embeddings=False, **kwargs):
        """
        Get AnnData object.

        """
        embeddings = None
        if add_embeddings and "emb_names" not in kwargs:
            embeddings = ["scvi", "geneformer"]
        else:
            embeddings = kwargs.get("emb_names")
        with cellxgene_census.open_soma(
                census_version=self.census_version) as census:
            adata = cellxgene_census.get_anndata(
                census=census,
                organism="Homo sapiens",
                var_value_filter=kwargs.get("var_value_filter"),
                obs_value_filter=kwargs.get("obs_value_filter"),
                obs_embeddings=embeddings,
            )
            return adata


if __name__ == "__main__":
    pass
